import pyspark.sql.functions as F
from pyspark.sql import DataFrame

from python_search.events.run_performed.dataset import EntryExecutedDataset


class TimesUsed:
    """
    Times used feature
    """

    def __init__(self):
        _df = EntryExecutedDataset().load_clean()
        self._df = (
            _df.groupBy("key")
            .agg(F.count("key").alias("times_used"))
            .sort("times_used", ascending=False)
        )
        self._pandas_df = self._df.toPandas()

    def get_dataframe(self) -> DataFrame:
        return self._df

    def get_value(self, key) -> int:

        result = self._pandas_df[self._pandas_df["key"] == key]

        if result is None or result.empty:
            return 0
        print(result)

        return result["times_used"].values[0]


if __name__ == "__main__":
    import fire

    fire.Fire(TimesUsed)



from __future__ import annotations

import logging
import threading
from dataclasses import dataclass
from typing import List

import fire

from python_search.config import ConfigurationLoader
from python_search.entry_type.classifier_inference import ClassifierInferenceClient
from python_search.infrastructure.arize import Arize
from python_search.sdk.web_api_sdk import PythonSearchWebAPISDK


class EntryCaptureGUI:
    def __init__(self):
        self._configuration = ConfigurationLoader().load_config()
        self._tags = self._configuration._default_tags
        self._prediction_uuid = None

    def launch(
        self,
        title: str = "Capture Entry",
        default_key: str = "",
        default_content: str = "",
        serialize_output=False,
        default_type="Snippet",
    ) -> GuiEntryData:
        """
        Launch the _entries capture GUI.
        """

        print("Default key: ", default_key)
        import PySimpleGUI as sg

        self._sg = sg

        config = ConfigurationLoader().load_config()
        sg.theme(config.simple_gui_theme)
        font_size = config.simple_gui_font_size

        content_input = sg.Input(
            key="content",
            default_text=default_content,
            expand_x=True,
            expand_y=True,
        )
        entry_type = sg.Combo(
            [
                "Snippet",
                "Cmd",
                "Url",
                "File",
                "Anonymous",
            ],
            key="type",
            default_value=default_type,
        )

        tags_chucks = self._chunks_of_tags(self._tags, 4)
        layout = [
            [sg.Text("Entry content")],
            [content_input],
            [sg.Text("Descriptive key name")],
            [
                sg.Input(
                    key="key", default_text=default_key, expand_x=True, expand_y=True
                )
            ],
            [sg.Text("Type")],
            [entry_type],
            [sg.Text("Tags")],
            [self._checkbox_list(i) for i in tags_chucks],
            [sg.Button("Write", key="write")],
        ]

        window = sg.Window(
            title,
            layout,
            font=("Helvetica", font_size),
            finalize=True,
        )

        # workaround for mac bug

        content_input.update(select=True)

        window["key"].bind("<Return>", "_Enter")
        window["content"].bind("<Return>", "_Enter")
        window["type"].bind("<Return>", "_Enter")
        window["key"].bind("<Escape>", "_Esc")
        window["content"].bind("<Escape>", "_Esc")
        window["type"].bind("<Escape>", "_Esc")

        threading.Thread(
            target=self._predict_entry_type, args=(window, default_content), daemon=True
        ).start()

        if not default_key:
            threading.Thread(
                target=self._generate_description,
                args=(window, default_content),
                daemon=True,
            ).start()

        while True:
            event, values = window.read()
            if event and (event == "write" or event.endswith("_Enter")):
                break

            if event == "-type-inference-ready-":
                window["type"].update(values[event])
                continue

            if event == "-generated-key-ready-":
                window["key"].update(values[event])
                continue

            if event == sg.WINDOW_CLOSED or event.endswith("_Esc"):
                raise Exception("Quitting window")

        window.close()
        logging.info("values", values)

        selected_tags = []
        for key, value in values.items():
            if key in self._tags and value == True:
                selected_tags.append(key)

        result = GuiEntryData(
            values["key"], values["content"], values["type"], selected_tags
        )

        self._report_actual(result)

        if serialize_output:
            result = result.__dict__

        return result

    def _report_actual(self, data: GuiEntryData):
        if not self._prediction_uuid:
            print("No prediction uuid, skipping report")
            return

        if not Arize.is_installed():
            return
        arize_client = Arize().get_client()

        from arize.utils.types import Environments, ModelTypes

        data = {
            "model_id": Arize.MODEL_ID,
            "model_version": Arize.MODEL_VERSION,
            "model_type": ModelTypes.SCORE_CATEGORICAL,
            "environment": Environments.PRODUCTION,
            "prediction_id": self._prediction_uuid,
            "actual_label": data.type,
        }
        print(f"Data to send arize: {data}")

        arize_result = arize_client.log(**data)
        Arize.arize_responses_helper(arize_result)

    def _predict_entry_type(self, window, content):
        result = ClassifierInferenceClient().predict_from_content(content)

        if not result:
            return

        new_type = result[0]
        self._prediction_uuid = result[1]
        print(f"New type: {new_type}, uuid: {self._prediction_uuid}")
        window.write_event_value("-type-inference-ready-", new_type)

    def _generate_description(self, window, content):
        result = PythonSearchWebAPISDK().generate_description(
            {"content": content, "temperature": 0.2}
        )

        if not result:
            return

        description = result["generated_description"]
        print(f"New description: {description}")
        window.write_event_value("-generated-key-ready-", description)

    def _checkbox_list(self, tags):
        return ([self._sg.Checkbox(tag, key=tag, default=False) for tag in tags],)

    def _chunks_of_tags(self, lst, n):
        """Yield successive n-sized chunks from lst."""
        for i in range(0, len(lst), n):
            yield lst[i : i + n]


@dataclass
class GuiEntryData:
    """
    Entry _entries schema

    """

    key: str
    value: str
    type: str
    tags: List[str]


if __name__ == "__main__":
    fire.Fire(EntryCaptureGUI().launch)



from __future__ import annotations

from python_search.events.run_performed import RunPerformed


class LogRunPerformedClient:
    def send(self, data: RunPerformed):
        import requests

        try:
            result = requests.post(
                url="http://localhost:8000/log_run", json=data.__dict__
            )
            return result
        except BaseException as e:
            print(f"Logging results failed, reason: {e}")


class RunPerformedWriter:
    """
    Writes event
    """

    def write(self, event: RunPerformed):
        import datetime

        event.timestamp = str(datetime.datetime.now(datetime.timezone.utc).timestamp())

        from python_search.data_collector import GenericDataCollector
        return GenericDataCollector().write(
            data=event.__dict__, table_name="searches_performed"
        )



from typing import Union

import requests
import json

from python_search.entry_description_generator.description_geneartor import (
    EntryKeyGeneratorCmd,
)


class PythonSearchWebAPISDK:
    """
    A lightweight SDK for the web api
    """

    def __init__(self):
        pass

    def recent_history(self, return_json=False):
        """
        Returns the recent history of the user

        """

        result = requests.get(url="http://localhost:8000/recent_history")
        if return_json:
            return result.text

        data = json.loads(result.text)
        return data["history"]

    def generate_description(
        self, generator_cmd: Union[dict, EntryKeyGeneratorCmd], return_json=False
    ):
        if type(generator_cmd) == dict:
            generator_cmd = EntryKeyGeneratorCmd(**generator_cmd)
        result = requests.post(
            url="http://localhost:8000/entry/generate_description",
            json=generator_cmd.dict(),
        )
        if return_json:
            return result.text

        data = json.loads(result.text)
        return data


def main():
    import fire

    fire.Fire(PythonSearchWebAPISDK)


if __name__ == "__main__":
    main()



class PythonSearchFeaturesSupport:
    """
    Enables us to have the core behaviour working without the latest in development features.

    """

    DYNAMIC_RANKING = "dynamic_ranking"

    DEFAULT_SUPPORT = {
        # if user history or is supported
        DYNAMIC_RANKING: False,
    }

    @staticmethod
    def default():
        return PythonSearchFeaturesSupport(PythonSearchFeaturesSupport.DEFAULT_SUPPORT)

    def __init__(self, config: dict):
        self.supported_config = {
            **PythonSearchFeaturesSupport.DEFAULT_SUPPORT,
            **config,
        }

    def is_enabled(self, feature_name) -> bool:
        if feature_name not in self.supported_config:
            raise Exception(f"Feature {feature_name} not configured")

        return self.supported_config[feature_name]

    def is_dynamic_ranking_supported(self):
        return self.is_enabled("dynamic_ranking")

    def is_redis_supported(self):
        return self.is_enabled("redis")



import os

from python_search.environment import is_mac


class Project:
    def new_project(self, new_project_location):
        """
        Initialize a new project to use Python search and make sure all remaining dependencies exist
        """

        if os.path.exists(new_project_location):
            raise Exception(f"{new_project_location} already exists")

        os.system(f"mkdir -p {new_project_location} 2>/dev/null")

        script_dir = os.path.dirname(os.path.realpath(__file__))
        copy_cmd = f"cp -r {script_dir}/entries_main.py {new_project_location}"
        os.system(copy_cmd)
        os.system(f"cd {new_project_location} && git init . 1>/dev/null ")

        self.set_current_project(new_project_location)

        print(
            f"""Project created successfully! 

Your main config script can be found at {new_project_location}/entries_main.py

You can now start using python search by issuing:
python_search search"""
        )

    def set_current_project(self, project_location: str):
        """
        Sets an existing project as the one python search is using.
        """
        import os

        home = os.environ["HOME"]
        current_project_config = f"{home}/.config/python_search/current_project"

        os.system(f"mkdir -p {home}/.config/python_search/")
        os.system(f'echo "{project_location}" >  {current_project_config}')

        print(f"Successfuly set current project as: {project_location}")



import uuid


from python_search.entry_type.classifier_inference import get_value_and_label
from python_search.entry_type.entity import EntryType
from python_search.config import ConfigurationLoader, DataConfig
import pandas as pd
import matplotlib.pyplot as plt

from python_search.infrastructure.arize import Arize
from python_search.search.next_item_predictor.mlflow_logger import configure_mlflow
from python_search.search.next_item_predictor.features.entry_embeddings.entry_embeddings import (
    create_embeddings_from_strings,
)
import os
import numpy as np
from keras import models
from keras import layers
from keras.utils.np_utils import to_categorical


class Pipeline:

    INPUT_DIMENSIONS = 384 + 2

    def __init__(self):
        self._arize_client = Arize().get_client()

    def run(self, embeddings_cache=None, cache=False, disable_arize=False):
        df = self.create_dataset()
        input_list = df["input"].tolist()
        number_of_items = len(input_list)
        embeddings = None

        # cache flag can enable caching of embeddings if not set specifically
        embeddings_cache = embeddings_cache if embeddings_cache else cache

        CACHE_LOCATION = "/root/.cache/entry_type_np_cache.npy"
        if embeddings_cache and os.path.exists(CACHE_LOCATION):
            print("Using transformed data from cache")
            embeddings = np.load(CACHE_LOCATION, allow_pickle=True)
            if len(embeddings) != number_of_items:
                print("Cache size does not match dataset size")
                embeddings = None

        if embeddings is None:
            embeddings = create_embeddings_from_strings(input_list)
            np.save(CACHE_LOCATION, embeddings)

        X = np.zeros([number_of_items, Pipeline.INPUT_DIMENSIONS])
        for i in range(0, number_of_items):
            X[i] = np.concatenate(
                (
                    embeddings[i],
                    np.asarray([1 if df["has_pipe"].iloc[i] else 0]),
                    np.asarray([1 if df["has_double_minus"].iloc[i] else 0]),
                )
            )

        categorical_labels = to_categorical(df["label"].tolist())

        mlflow = configure_mlflow(
            experiment_name=DataConfig.ENTRY_TYPE_CLASSIFIER_EXPERIMENT_NAME
        )
        mlflow.keras.autolog()
        with mlflow.start_run() as modelInfo:
            model = models.Sequential()
            run_id = mlflow.active_run().info.run_id
            model.add(
                layers.Dense(
                    128, activation="relu", input_shape=(Pipeline.INPUT_DIMENSIONS,)
                )
            )
            model.add(layers.Dense(64, activation="relu"))
            model.add(layers.Dense(5, activation="softmax"))

            model.compile(
                optimizer="rmsprop",
                loss="categorical_crossentropy",
                metrics=["accuracy"],
            )

            x_train = X[1000:]
            y_train = categorical_labels[1000:]

            # separate 1000 for validation
            x_val = X[:1000]
            y_val = categorical_labels[:1000]

            history = model.fit(
                x_train,
                y_train,
                epochs=20,
                batch_size=512,
                validation_data=(x_val, y_val),
            )
            mlflow.keras.log_model(model, "model", keras_module="keras")
            run_id = mlflow.active_run().info.run_id
            print("run id: ", run_id)

        if not disable_arize and Arize.is_installed():
            from arize.utils.types import ModelTypes, Environments, Embedding

            print("Sending validation values to arize")
            for key, entry in enumerate(x_val):
                value, label = get_value_and_label(
                    model.predict(np.asarray([entry]))[0]
                )
                result = self._arize_client.log(
                    model_id=Arize.MODEL_ID,
                    model_version=Arize.MODEL_VERSION,
                    model_type=ModelTypes.SCORE_CATEGORICAL,
                    batch_id=run_id,
                    prediction_id=str(uuid.uuid4()),
                    features={"has_pipe": entry[384], "has_double_minus": entry[385]},
                    embedding_features={
                        "content": Embedding(vector=entry, data=input_list[key])
                    },
                    prediction_label=(label, float(value)),
                    actual_label=(
                        get_value_and_label(y_val[key])[1],
                        float(get_value_and_label(y_val[key])[0]),
                    ),
                    environment=Environments.VALIDATION,
                )
                Arize.arize_responses_helper(result)

        loss = history.history["loss"]
        val_loss = history.history["val_loss"]

        epochs = range(1, len(loss) + 1)
        plt.plot(epochs, loss, "bo", label="TrainingLoss")
        plt.plot(epochs, val_loss, "b", label="ValidationLoss")
        plt.legend()
        plt.show()

        accuracy = history.history["accuracy"]
        val_acc = history.history["val_accuracy"]

        epochs = range(1, len(loss) + 1)
        plt.plot(epochs, accuracy, "bo", label="TrainingAccuracy")
        plt.plot(epochs, val_acc, "b", label="ValidationAccuracy")
        plt.legend()
        plt.show()

    def create_dataset(self) -> pd.DataFrame:
        entries = ConfigurationLoader().load_entries()
        entries
        dataset = []

        for key, entry in entries.items():
            X = key + " "
            y = None
            if "url" in entry:
                y = EntryType.to_categorical(EntryType.URL)
                X = X + str(entry["url"])
            if "snippet" in entry:
                y = EntryType.to_categorical(EntryType.SNIPPET)
                X = X + str(entry["snippet"])
            if "file" in entry:
                y = EntryType.to_categorical(EntryType.FILE)
                X = X + str(entry["file"])
            if "callable" in entry:
                y = EntryType.to_categorical(EntryType.CALLABLE)
                X = X + str(entry["callable"])
            if "cmd" in entry:
                y = EntryType.to_categorical(EntryType.CMD)
                X = X + str(entry["cmd"])
            if "cli_cmd" in entry:
                y = EntryType.to_categorical(EntryType.CMD)
                X = X + str(entry["cli_cmd"])

            if y is None:
                raise Exception(f"Could not map key '{key}' to any value {entry}")

            has_pipe = "|" in X
            has_double_minus = "--" in X

            dataset.append([X, has_pipe, has_double_minus, y])

        df = pd.DataFrame(
            dataset, columns=["input", "has_pipe", "has_double_minus", "label"]
        )
        return df


if __name__ == "__main__":
    import fire

    fire.Fire()



from __future__ import annotations

from typing import List, Tuple


class RankedEntries:
    type = List[Tuple[str, dict]]

    @staticmethod
    def get_list_of_tuples(from_keys: List[str], entities: dict) -> RankedEntries.type:
        used_entries = []
        for used_key in from_keys:
            if used_key not in entities:
                continue
            used_entries.append((used_key, entities[used_key]))
        return used_entries



import os


class Context:
    """
    captures information from the environment
    to contextualize how to execute the command
    """

    _instance = None
    _is_cli: bool = False
    _is_group_command = False
    _should_execute_sequentially = False
    _input = None

    @staticmethod
    def get_instance():
        if not Context._instance:
            Context._instance = Context()

        return Context._instance

    def is_cli(self):
        if os.environ.get("DISABLE_CLI"):
            return False

        return self._is_cli

    def enable_gui_mode(self):
        self._is_cli = False
        return self

    def enable_group_command(self):
        self._is_group_command = True
        return self

    def is_group_command(self):
        return self._is_group_command

    def enable_sequential_execution(self):
        self._should_execute_sequentially = True

    def disable_sequential_execution(self):
        self._should_execute_sequentially = False

    def should_execute_sequentially(self):
        return self._should_execute_sequentially

    def set_interpreter(self, interpreter):
        self._interpreter = interpreter

    def get_interpreter(self):
        return self._interpreter

    def set_input(self, input):
        self._input = input

    def get_input(self):
        return self._input



from __future__ import annotations

import inspect
import logging
import os


class EntriesGroup:
    """
    Main _configuration of the application. Customers are supposed to pass their own
    """

    # the location of the dumped index
    commands: dict = {}

    def aggregate_commands(self, commands_classes):
        """
        aggregates a list of classes or instances
        """
        for class_i in commands_classes:
            is_class = inspect.isclass(class_i)
            instance = class_i() if is_class else class_i

            if isinstance(instance, EntriesGroup):
                cmd_items = instance.get_hydrated_commands()
            else:
                cmd_items = instance.commands

            self.commands = {**self.commands, **cmd_items}

    def get_command(self, given_key):
        """Returns command value based on the key name, must match 11"""
        given_key = given_key.lower()
        for key, value in self.commands.items():
            if key.lower() == given_key:
                return value

        raise Exception(f"Value not found for key: {given_key}")

    def get_keys(self):
        keys = []
        for key, value in self.commands.items():
            keys.append(key)

        return keys

    def get_hydrated_commands(self):
        result = {}
        class_name_tag = self.__class__.__name__
        for key, command in self.commands.items():
            if type(command) is dict:
                if "tags" not in command:
                    command["tags"] = [class_name_tag]
                else:
                    if class_name_tag not in command["tags"]:
                        command["tags"].append(class_name_tag)

            result[key] = command

        return result

    def get_source_file(self):
        """
        Returns the path of the source code where the config is stored
        """
        import sys

        return sys.argv[0]

    def get_project_root(self):
        """
        @ deprecated use environment loader instead
        """

        from python_search.config import ConfigurationLoader

        return ConfigurationLoader().get_project_root()



import logging
import os


class FeatureToggle:
    """
    A simple feature toggle abstraction that allows one to turn on/off features for their environment without touching code.
    """

    @staticmethod
    def main():
        import fire

        fire.Fire(FeatureToggle)

    def __init__(self):
        self.BASE_PATH = f"{os.getenv('HOME')}/.PythonSearch/features"

        if not os.path.exists(self.BASE_PATH):
            os.system(f"mkdir -p '{self.BASE_PATH}'")

    def enable(self, feature_name: str):
        os.system(f"touch {self.BASE_PATH}/{feature_name}")

    def disable(self, feature_name: str):
        os.system(f"rm {self.BASE_PATH}/{feature_name}")

    def toggle(self, feature_name):
        is_enabled = self.is_enabled(feature_name)
        print(f"feature: {feature_name} is enabled={is_enabled}, inverting it")

        if is_enabled:
            self.disable(feature_name)
        else:
            self.enable(feature_name)

    def is_enabled(self, feature_name: str) -> bool:
        result = 0 == os.system(f" test -f {self.BASE_PATH}/{feature_name}")

        if result:
            logging.debug(f"Feature toggle {feature_name} is enabled")
        else:
            logging.debug(f"Feature toggle {feature_name} is disabled")

        return result



from __future__ import annotations

from pyspark.sql.types import StructType, StructField, StringType, IntegerType

from python_search.config import DataConfig
from python_search.data_collector import GenericDataCollector


class EntryExecutedDataset:
    """
    Poit of access for the searches performed
    This is the place to keep the source of truth for the schema as well
    """

    columns = ["key", "query_input", "shortcut", "rank_uuid", "timestamp"]
    TABLE_NAME = "run_performed"
    NEW_TABLE_NAME = "searches_performed"
    CLEAN_PATH = DataConfig.CLEAN_EVENTS_FOLDER + "/" + TABLE_NAME

    def __init__(self, spark=None):
        from pyspark.sql.session import SparkSession

        self.spark = spark if spark else SparkSession.builder.getOrCreate()

    def load_old(self):
        """
        Load old original events stored via spark-streaming
        """
        data_location = "file://" + DataConfig.SEARCH_RUNS_PERFORMED_FOLDER
        print("Loading data from: {}".format(data_location))
        return self.spark.read.format("parquet").load_old(data_location)

    def load_clean(self):
        return self.spark.read.parquet(self.CLEAN_PATH)

    def load_new(self):
        from pyspark.sql.session import SparkSession

        schema = StructType(
            [
                StructField("key", StringType(), True),
                StructField("query_input", StringType(), True),
                StructField("shortcut", StringType(), True),
                StructField("rank_uuid", StringType(), True),
                StructField("rank_position", IntegerType(), True),
                StructField("timestamp", StringType(), True),
            ]
        )

        spark = SparkSession.builder.getOrCreate()
        result_df = spark.read.json(
            GenericDataCollector().data_location(EntryExecutedDataset.NEW_TABLE_NAME),
            schema=schema,
        )

        return result_df



"""
Clients should depend on a configuration instance (config) rather than in the class,
the class should only be used for type annotation.
This way we can have multiple configs depending of the enviroment.
"""
import datetime
import logging
import os
from typing import List, Optional, Tuple

from python_search.entries_group import EntriesGroup
from python_search.environment import is_mac
from python_search.features import PythonSearchFeaturesSupport




class DataConfig:
    """
    Configuration with locations of _entries, models and names of files.
    """

    # output of the _model
    base_data_folder = f"{os.environ['HOME']}/.data/data/python_search"
    BASE_DATA_COLLECTOR_FOLDER = f"{os.environ['HOME']}/.data/data_collection"
    prediction_batch_location = base_data_folder + "/predict_input_lenght/latest"
    # a copy of the search run _entries for the feature store
    entries_dump = base_data_folder + "/entries_dumped/latest"
    entries_dump_file = base_data_folder + "/entries_dumped/latest/000.parquet"
    cached_configuration = "/tmp/search_and_run_configuration_cached"
    MLFLOW_MODELS_PATH = f"/entries/mlflow"
    BASELINE_EXPERIMENT_NAME = f"baseline_rank_v0"
    NEXT_ITEM_EXPERIMENT_NAME = f"next_item_v1"
    ENTRY_TYPE_CLASSIFIER_EXPERIMENT_NAME = f"entry_type_classifier_v2"
    DATA_WAREHOUSE_FOLDER = base_data_folder + "/data_warehouse"
    CLEAN_EVENTS_FOLDER = DATA_WAREHOUSE_FOLDER + "/clean"
    SEARCH_RUNS_PERFORMED_FOLDER = (
        base_data_folder + "/data_warehouse/dataframes/SearchRunPerformed"
    )


# @todo do not depend on this config directly rather depend on the base configuration

class KafkaConfig:
    default_port: str = "9092"
    host: str = f"127.0.0.1:{default_port}"


class RedisConfig:
    host = "127.0.0.1" if is_mac() else "host.docker.internal"
    port = 6379


class PythonSearchConfiguration(EntriesGroup):
    """
    The main _configuration of Python Search
    Everything to customize about the application is configurable via code through this class
    """

    APPLICATION_TITLE = "PythonSearchWindow"
    commands: dict
    simple_gui_theme = "SystemDefault1"
    simple_gui_font_size = 14
    _default_tags = None
    tags_dependent_inserter_marks = None
    _initialization_time = None
    _default_text_editor = "vim"

    def __init__(
        self,
        *,
        entries: Optional[dict] = None,
        entries_groups: Optional[List[EntriesGroup]] = None,
        supported_features: Optional[PythonSearchFeaturesSupport] = None,
        default_tags=None,
        tags_dependent_inserter_marks: Optional[dict[str, Tuple[str, str]]] = None,
        default_text_editor: Optional[str] = None,
    ):
        """

        :param entries:
        :param entries_groups:
        :param supported_features:
        :param default_tags:
        :param tags_dependent_inserter_marks:
        :param default_text_editor:
        """
        if entries:
            self.commands = entries

        if entries_groups:
            self.aggregate_commands(entries_groups)

        if supported_features:
            self.supported_features: PythonSearchFeaturesSupport = supported_features
        else:
            self.supported_features: PythonSearchFeaturesSupport = (
                PythonSearchFeaturesSupport.default()
            )

        if default_tags:
            self._default_tags = default_tags

        self.tags_dependent_inserter_marks = tags_dependent_inserter_marks


        self._initialization_time = datetime.datetime.now()
        self._default_text_editor = default_text_editor


    def get_text_editor(self):
        return self._default_text_editor

    def get_default_tags(self):
        return self._default_tags


class ConfigurationLoader:
    """
    Loads the application from the environment.py
    """

    def load_config(self) -> PythonSearchConfiguration:

        folder = self.get_project_root()
        config_location = f"{folder}/entries_main.py"

        if not os.path.exists(config_location):
            raise Exception(f"Could not find config file {config_location}")

        import sys

        if folder not in sys.path:
            sys.path.insert(0, folder)
        import copy
        from entries_main import config

        return copy.deepcopy(config)

    def reload(self):
        """
        reload _entries for when we change it
        """
        import importlib

        import entries_main

        importlib.reload(entries_main)

        import copy

        import entries_main

        return copy.deepcopy(entries_main.config)

    def get_project_root(self):
        env_name = "PS_ENTRIES_HOME"
        current_project_location = (
            os.environ["HOME"] + "/.config/python_search/current_project"
        )

        folder = None

        if env_name in os.environ:
            logging.debug(
                f"Env exists and takes precedence: {env_name}={os.environ[env_name]}"
            )
            folder = os.environ[env_name]

        if os.path.isfile(current_project_location):
            with open(current_project_location) as f:
                folder = f.readlines()[0].strip()

        if not folder:
            raise Exception(
                f"Either {current_project_location} or {env_name} must be set to find _entries"
            )
        return folder

    def load_entries(self):
        return self.load_config().commands



#!/usr/bin/env python
from python_search.config import RedisConfig


class PythonSearchRedis:
    @staticmethod
    def get_client():
        import redis

        return redis.StrictRedis(host=RedisConfig.host, port=RedisConfig.port)


class RedisCli:
    def cli(self):
        import os

        os.system(f"redis-cli -h {RedisConfig.host} -p {RedisConfig.port}")


if __name__ == "__main__":
    import fire

    fire.Fire(RedisCli)



import os
import time

from python_search.apps.terminal import Terminal


class AskQuestion:
    def ask(self, message: str) -> str:

        message = self._sanitize_message(message)
        content_file = "/tmp/python_search_input"
        if os.path.exists(content_file):
            os.remove(content_file)

        cmd = f"""kitty {Terminal.GENERIC_TERMINAL_PARAMS} bash -c 'printf "{message}: "; read tmp; echo "$tmp" >{content_file}' &"""
        os.system(cmd)

        while not os.path.exists(content_file):
            time.sleep(1)

        time.sleep(0.2)

        with open(content_file) as file:
            result = file.readlines()[0].strip()

        return result

    def _sanitize_message(self, message):
        allowed_chars = [" ", ":", "\n"]
        return "".join(
            char for char in message if char.isalnum() or char in allowed_chars
        )



import pyspark.sql.functions as F


class RunPerformedCleaning:
    def clean(self):
        print("Performing the cleaning of the new events")
        from python_search.events.run_performed.dataset import EntryExecutedDataset

        # load baseline clean
        df_clean = EntryExecutedDataset().load_clean()
        print(f"Number of pre-existing clean events: {df_clean.count()}")

        # get  latest timestamp imported
        max_timestamp = df_clean.agg({"timestamp": "max"}).collect()[0][0]

        df_new = EntryExecutedDataset().load_new().sort("timestamp", ascending=False)
        df_new = df_new.withColumn(
            "timestamp_real", F.from_unixtime(F.col("timestamp"))
        )
        df_new = df_new.filter("timestamp IS NOT NULL").filter(
            f"timestamp_real > '{max_timestamp}'"
        )
        df_new = df_new.drop("timestamp")
        df_new = df_new.withColumnRenamed("timestamp_real", "timestamp")
        df_new = df_new.withColumn("date", F.to_date("timestamp"))

        print(f"Number of new events: {df_new.count()}")

        joined = df_clean.union(df_new)
        print("Joined size", joined.count())

        joined.write.option("partitionOverwriteMode", "dynamic").partitionBy(
            "date"
        ).mode("overwrite").parquet(EntryExecutedDataset.CLEAN_PATH)

        df_clean = EntryExecutedDataset().load_clean()
        print("Number of clean events", df_clean.count())


if __name__ == "__main__":
    RunPerformedCleaning().clean()



import logging

from python_search.apps.window_manager import WindowManager
from python_search.config import ConfigurationLoader, PythonSearchConfiguration
from python_search.entry_runner import EntryRunner
from python_search.environment import is_mac
from python_search.events.run_performed import RunPerformed
from python_search.events.run_performed.writer import LogRunPerformedClient
from python_search.search_ui.fzf_kitty import FzfInKitty
from python_search.search_ui.preview import Preview
from python_search.exceptions import notify_exception


class PythonSearchCli:
    """
    The command line application, entry point of the program.

    Try to avoid adding direct commands, prefer instead to add objects as parts of functions
    """

    # all commands that are not self-explanatory should not be part of the main api, thus are marked as private.

    configuration: PythonSearchConfiguration

    @staticmethod
    def install_dependencies():
        """install dependenceis"""
        from python_search.init.install_dependencies import InstallDependencies

        InstallDependencies().install_all()

    @staticmethod
    def new_project(project_name: str):
        """Create a new project in the current directory with the given name"""
        from python_search.init.project import Project

        Project().new_project(project_name)

    @staticmethod
    def set_project_location(location: str):
        """Create a new project in the current directory with the given name"""
        from python_search.init.project import Project

        Project().set_current_project(location)

    def __init__(self, configuration: PythonSearchConfiguration = None):
        """
        Keep this constructor small and import dependencies inside the functions
        so they keep being fast
        """
        if not configuration:
            logging.debug("No _configuration provided, using default")
            try:
                configuration = ConfigurationLoader().load_config()
            except Exception as e:
                print(f"Did not find any config to load, error: {e}")

                # early return so we dont set the remaining attributes
                return

        self.configuration = configuration
        self.run_key = EntryRunner(self.configuration).run

    def search(self):
        """
        Opens the Search UI. Main entrypoint of the application
        """
        FzfInKitty(self.configuration).run()

    @notify_exception()
    def edit_main(self):
        """Edit the main script"""
        from python_search.entry_capture.edit_content import EditKey

        return EditKey(self.configuration).edit_default()

    def register_new(self):
        """Starts the UI for collecting a new entry into pythonsearch"""
        from python_search.entry_capture.register_new import RegisterNew

        return RegisterNew(self.configuration)

    def edit_key(self, entry_str):
        """Opens the key in the source code using the IDE specified in the config, defaults to vim"""
        from python_search.entry_capture.edit_content import EditKey

        result = EditKey(self.configuration).edit_key(entry_str, dry_run=False)
        key = entry_str.split(":")[0]
        LogRunPerformedClient().send(
            RunPerformed(key=key, query_input="", shortcut=False)
        )
        return result

    def _copy_entry_content(self, entry_str: str):
        """
        Copies the content of the provided key to the clipboard.
        Used by fzf to provide Ctrl-c functionality.
        """
        from python_search.interpreter.interpreter_matcher import InterpreterMatcher

        key = entry_str.split(":")[0]
        InterpreterMatcher.build_instance(self.configuration).clipboard(key)
        print(f"Key: {key}")
        LogRunPerformedClient().send(
            RunPerformed(key=key, query_input="", shortcut=False)
        )

    def _copy_key_only(self, entry_str: str):
        """
        Copies to clipboard the key
        """
        from python_search.apps.clipboard import Clipboard

        key = entry_str.split(":")[0]
        Clipboard().set_content(key, enable_notifications=True)
        LogRunPerformedClient().send(
            RunPerformed(key=key, query_input="", shortcut=False)
        )

    def shortcut(self):
        """Generate shorcuts for all environments"""
        from python_search.shortcut.generator import ShortcutGenerator

        return ShortcutGenerator(self.configuration)

    def _search_edit(self, entry_str=None):
        from python_search.entry_capture.edit_content import EditKey

        result = EditKey(self.configuration).search_entries_directory(entry_str)

        if entry_str.split(":"):
            key = entry_str.split(":")[0]
            LogRunPerformedClient().send(
                RunPerformed(key=key, query_input="", shortcut=False)
            )
        return result

    def _ranking(self):
        from python_search.search.search import Search

        return Search(self.configuration)

    def _features(self):
        """Feature toggle system"""
        from python_search.feature_toggle import FeatureToggle

        return FeatureToggle()

    def _utils(self):
        """Here commands that are small topics and dont fit the rest"""

        class Utils:
            def __init__(self, configuration):
                self.configuration = configuration

            def hide_launcher(self):
                """hide the search launcher -i2 specific"""
                if is_mac():
                    import os

                    os.system(
                        """osascript -e 'tell application "System Events" to keystroke "H" using {command down}'"""
                    )
                WindowManager.load_from_environment().hide_window(
                    self.configuration.APPLICATION_TITLE
                )

        return Utils(self.configuration)

    def _preview_entry(self, entry_text: str):
        """
        Recieves _entries from fzf and show them formatted for the preview window
        """
        Preview().display(entry_text)

    def _infra_report(self):
        from python_search.infrastructure.report import Report

        return Report()

    def _entry_type_classifier(self):

        from python_search.entry_type.classifier_inference import (
            ClassifierInferenceClient,
        )

        class EntryTypeClassifierAPI:
            def __init__(self):
                self.inference_client = ClassifierInferenceClient

        return EntryTypeClassifierAPI


def _error_handler(e):
    import sys
    import traceback

    exc_info = sys.exc_info()
    logging.warning(
        f"Unhandled exception: {e}".join(traceback.format_exception(*exc_info))
    )

    raise e


def _run_key_bin():
    """
    Entry point to run a key
    """
    import fire

    fire.Fire(PythonSearchCli().run_key)


def main():
    import fire

    fire.Fire(PythonSearchCli)



from __future__ import annotations

import datetime
import os
from typing import Optional

from python_search.search.next_item_predictor.features.times_used import TimesUsed


class ModelInput:
    hour: int
    month: int
    previous_key: str
    previous_previous_key: str
    times_used_previous: int
    times_used_previous_previous: int

    _times_used: Optional[TimesUsed] = None

    def __init__(
        self,
        *,
        hour,
        month,
        previous_key,
        previous_previous_key,
        times_used_previous: Optional[int] = None,
        times_used_previous_previous: Optional[int] = None,
    ):
        self.hour = hour
        self.month = month
        self.previous_key = previous_key
        self.previous_previous_key = previous_previous_key
        self.times_used_previous = times_used_previous
        self.times_used_previous_previous = times_used_previous_previous

        if not ModelInput._times_used:
            ModelInput._times_used = TimesUsed()

        if not times_used_previous:
            self.times_used_previous = ModelInput._times_used.get_value(previous_key)
        if not times_used_previous_previous:
            self.times_used_previous_previous = ModelInput._times_used.get_value(
                previous_previous_key
            )

    @staticmethod
    def with_given_keys(previous_key: str, previous_previous_key: str) -> "ModelInput":
        """
        Do inference based on the current time and the recent used keys
        """
        now = datetime.datetime.now()
        hour = int(os.environ["FORCE_HOUR"]) if "FORCE_HOUR" in os.environ else now.hour
        month = (
            int(os.environ["FORCE_MONTH"]) if "FORCE_MONTH" in os.environ else now.month
        )

        instance = ModelInput(
            hour=hour,
            month=month,
            previous_key=previous_key,
            previous_previous_key=previous_previous_key,
        )

        print("Inference input: ", instance.__dict__)

        return instance



from __future__ import annotations

import datetime
import time

from python_search.apps.clipboard import Clipboard
from python_search.config import ConfigurationLoader
from python_search.entry_capture.filesystem_entry_inserter import (
    FilesystemEntryInserter,
)
from python_search.entry_capture.entry_inserter_gui import EntryCaptureGUI, GuiEntryData
from python_search.entry_type.entity import infer_default_type
from python_search.exceptions import RegisterNewException
from python_search.interpreter.base import BaseInterpreter
from python_search.interpreter.interpreter_matcher import InterpreterMatcher
from python_search.exceptions import notify_exception


class RegisterNew:
    """
    Entry point for the registering of new entries
    """

    def __init__(self, configuration=None):

        if not configuration:
            configuration = ConfigurationLoader().load_config()

        self.configuration = configuration
        self.entry_inserter = FilesystemEntryInserter(configuration)

    @notify_exception()
    def register(self, *, key: str, value: str, tag: str = None):
        """
        The non ui driven registering api
        Args:
            key:
            value:
            tag:

        Returns:

        """
        print(f"Registering new entry with tag = {tag}")
        key = self._sanitize_key(key)

        interpreter: BaseInterpreter = InterpreterMatcher.build_instance(
            self.configuration
        ).get_interpreter(value)
        dict_entry = interpreter.to_dict()
        if tag:
            dict_entry["tags"] = [tag]

        self.entry_inserter.insert(key, dict_entry)

    def _sanitize_key(self, key):
        return key.replace("\n", " ").replace(":", " ")

    @notify_exception()
    def launch_ui(self, default_type=None, default_key=None, default_content=None):
        """
        Create a new inferred entry based on the clipboard content
        """

        if not default_content:
            default_content = Clipboard().get_content()

        # @todo reenable this once in the dockerfile
        if not default_type:
            default_type = infer_default_type(default_content)

        entry_data: GuiEntryData = EntryCaptureGUI().launch(
            "New Entry Details",
            default_content=default_content,
            default_key=default_key,
            default_type=default_type,
        )
        key = self._sanitize_key(entry_data.key)
        interpreter: BaseInterpreter = InterpreterMatcher.build_instance(
            self.configuration
        ).get_interpreter_from_type(entry_data.type)

        dict_entry = interpreter(entry_data.value).to_dict()
        if entry_data.tags:
            dict_entry["tags"] = entry_data.tags

        self.entry_inserter.insert(key, dict_entry)

    def anonymous_snippet(self):
        """
        Create an anonymous snippet entry
        """

        now = datetime.datetime.now()
        key = f"no key {now.strftime('%Y %M %d %H %M %S')}"

        self.launch_ui(default_key=key, default_type="Snippet")

    def german_from_text(self, german_term: str):
        """
        @todo move this out of here to a plugin system
        Register german workds you dont know by saving them to the clipboard and storing in python search
        """

        if len(german_term) == 0:
            raise RegisterNewException.empty_content()

        from python_search.interpreter.url import UrlInterpreter

        print(f"german term: {german_term}")
        cmd = {
            "url": f"https://translate.google.com/?sl=de&tl=en&text={german_term}&op=translate"
        }
        UrlInterpreter(cmd).interpret_default()
        time.sleep(1)

        from python_search.apps.collect_input import CollectInput

        meaning = CollectInput().launch(f"Please type the meaning of ({german_term})")

        if not meaning:
            raise RegisterNewException.empty_content()

        as_dict = {
            "snippet": meaning,
            "language": "German",
        }

        self.entry_inserter.insert(german_term, as_dict)



from python_search.context import Context
from python_search.exceptions import CommandDoNotMatchException
from python_search.interpreter.base import BaseInterpreter


class GroupInterpreter(BaseInterpreter):
    def __init__(self, cmd, context: Context):
        self.context = context
        self.cmd = {}

        if type(cmd) is dict and "members" in cmd:
            if cmd["members"] is dict:
                raise Exception(
                    f"Members as dict are no longer valid, use list instead"
                )

            self.cmd["members"] = cmd["members"]
            self.context.enable_group_command()
            if "sequential" in cmd:
                self.context.enable_sequential_execution()

            return

        raise CommandDoNotMatchException(f"Not Valid members command {cmd}")

    def interpret_default(self):
        for member_key in self.cmd["members"]:
            self.context.get_interpreter().default(member_key)

        return



import os

from python_search.environment import is_mac


class InstallDependencies:
    def install_all(self):
        """
        Install all depenenceis to make python search work
        """

        self._install_fzf()
        self._install_kitty()

    def _install_fzf(self):
        print("Installing FZF")

        if self._exists("fzf"):
            return

        print("Looks like kitty is not installed in your platform. ")
        if is_mac():
            print("Installing it for you...")

            if not self._exists("brew"):
                raise Exception(
                    "Cannot continue without brew to install fzf. Please install brew first (https://brew.sh/)"
                )

            os.system("brew install fzf")
        else:
            print(
                "Dont know how to install fzf for your platform, please do so manually"
            )

    def _exists(self, cmd: str):
        result = os.system(f"which {cmd} >/dev/null")
        if result == 0:
            print(f"Great, you have {cmd} alread installed")
            return True
        return False

    def _install_kitty(self):
        if self._exists("kitty"):
            return

        print(
            "Looks like kitty is not installed in your platform. Installing it for you..."
        )
        os.system(
            "curl -L https://sw.kovidgoyal.net/kitty/installer.sh | sh /dev/stdin"
        )

        # sets a light theme in kitty
        os.system("kitty +kitten themes --reload-in=all One Half Light")



#!/usr/bin/env python
# using ipython interferes with fire arguments passing

import os
from typing import List, Literal, Optional

import numpy as np
from pyspark.sql import DataFrame, SparkSession
from sklearn.model_selection import train_test_split

from python_search.events.run_performed.clean import RunPerformedCleaning
from python_search.search.next_item_predictor.mlflow_logger import configure_mlflow
from python_search.search.next_item_predictor.offline_evaluation import (
    OfflineEvaluation,
)
from python_search.search.next_item_predictor.train_xgboost import TrainXGBoost
from python_search.search.next_item_predictor.training_dataset import TrainingDataset


class NextItemPredictorPipeline:
    """
    Exposes the whole ML pipeline, the runs everything
    """

    model_types: Literal["xgboost", "keras"] = "xgboost"

    def __init__(self):
        os.environ["TIME_IT"] = "1"

    def run(
        self,
        train_only: Optional[List[model_types]] = None,
        use_cache=False,
        clean_first=True,
        skip_offline_evaluation=False,
    ):
        """
        Trains both xgboost and keras models

        Args:
            train_only:
            use_cache:
            clean_first:

        Returns:

        """

        print("Using cache is:", use_cache)

        if not train_only:
            train_only = ["xgboost", "keras"]
        else:
            print("Training only: ", train_only)

        if clean_first:
            RunPerformedCleaning().clean()

        dataset: DataFrame = TrainingDataset().build(use_cache)
        from python_search.search.next_item_predictor.transform import ModelTransform

        X, Y = ModelTransform().transform_collection(dataset, use_cache=use_cache)
        from python_search.search.next_item_predictor.train_keras import TrainKeras

        X_train, X_test, Y_train, Y_test = train_test_split(
            X, Y, test_size=TrainKeras.TEST_SPLIT_SIZE, random_state=42
        )

        X_test_p = X_test
        X_test = np.delete(X_test, 0, axis=1)
        X_train = np.delete(X_train, 0, axis=1)

        if "xgboost" in train_only:
            mlflow = configure_mlflow()
            with mlflow.start_run():
                model = TrainXGBoost().train(X_train, X_test, Y_train, Y_test)
                mlflow.xgboost.log_model(model, "model")
                if not skip_offline_evaluation:
                    offline_evaluation = OfflineEvaluation().run(
                        model, dataset, X_test_p
                    )
                    mlflow.log_params(offline_evaluation)

        if "keras" in train_only:
            mlflow = configure_mlflow()
            with mlflow.start_run():
                from python_search.search.next_item_predictor.train_keras import (
                    TrainKeras,
                )

                model = TrainKeras().train(X_train, X_test, Y_train, Y_test)
                mlflow.keras.log_model(model, "model", keras_module="keras")
                if not skip_offline_evaluation:
                    offline_evaluation = OfflineEvaluation().run(
                        model, dataset, X_test_p
                    )
                    mlflow.log_params(offline_evaluation)

    def _spark(self):
        return SparkSession.builder.getOrCreate()


def main():
    import fire

    fire.Fire(NextItemPredictorPipeline)


if __name__ == "__main__":
    main()



import sys
import fire

from python_search.apps.clipboard import Clipboard


class CollectInput:
    """
    Ask the user for input and return the entered _entries
    """

    def launch(
        self,
        name="Enter Data",
        default_content="",
        prefill_with_clipboard: bool = False,
    ):
        """
        Launch the _entries capture GUI.
        """
        import contextlib

        with contextlib.redirect_stdout(None):
            import PySimpleGUI as sg

            if prefill_with_clipboard:
                default_content = Clipboard().get_content()

            font_size = 14
            sg.theme("SystemDefault1")

            input_field = sg.Input(
                key="content",
                default_text=default_content,
                expand_x=True,
                expand_y=True,
            )

            layout = [
                [input_field],
                [sg.Button("Continue", key="write")],
            ]

            window = sg.Window(
                name,
                layout,
                finalize=True,
                font=("Helvetica", font_size),
                alpha_channel=0.99,
            )

            if default_content != "":
                input_field.update(select=True)
            # workaround for mac bug
            window.read(timeout=100)
            window.set_alpha(1.0)

            window["content"].bind("<Return>", "_Enter")
            window["content"].bind("<Escape>", "_Esc")

            while True:
                event, values = window.read()

                if event and (event == "write" or event.endswith("_Enter")):
                    break
                if event == sg.WINDOW_CLOSED or event.endswith("_Esc"):
                    sys.exit(1)

            window.close()
        return values["content"]


def main():
    fire.Fire(CollectInput().launch)


if __name__ == "__main__":
    main()



def load_results_page():
    import pandas as pd
    import streamlit as st

    from python_search.config import ConfigurationLoader

    config = ConfigurationLoader().load_config()

    keys = config.commands.keys()

    from python_search.search.next_item_predictor.inference.inference import Inference
    from python_search.search.next_item_predictor.inference.input import ModelInput

    st.write("### Prediction results")
    st.write("##### Production run: " + Inference.PRODUCTION_RUN_ID)

    scenarios = {
        "work vs non work typical hours": {
            "a": {
                "previous_key": "ml platform roadmap team event session",
                "previous_previous_key": "ml platform teams spotify",
                "hour": 14,
                "month": 6,
            },
            "b": {
                "previous_key": "but tickets to brazil",
                "previous_previous_key": "ml platform teams spotify",
                "hour": 21,
                "month": 6,
            },
        },
        "work vs non work typical hours 2": {
            "a": {
                "previous_key": "git log in catalog",
                "previous_previous_key": "ml platform teams spotify",
                "hour": 14,
                "month": 6,
            },
            "b": {
                "previous_key": "precos agosto enviado ao pai",
                "previous_previous_key": "ml platform teams spotify",
                "hour": 21,
                "month": 6,
            },
        },
        "work in and not business hours": {
            "a": {
                "previous_key": "ml platform roadmap team event session",
                "previous_previous_key": "ml platform teams spotify",
                "hour": 11,
                "month": 6,
            },
            "b": {
                "previous_key": "ml platform roadmap team event session",
                "previous_previous_key": "ml platform teams spotify",
                "hour": 23,
                "month": 6,
            },
        },
    }
    current_scenario = st.selectbox("Scenario", scenarios)

    colA, colB = st.columns(2)

    def get_inference_input_for_scenario(scenario, a_or_b="a") -> ModelInput:
        input_data = scenarios[scenario][a_or_b]
        return ModelInput(**input_data)

    def perform_inference(inference_input):
        results = Inference(configuration=config).get_ranking(
            inference_input, return_weights=True
        )
        return pd.DataFrame.from_dict(results)

    colA.write("##### Input A")
    inference_a = get_inference_input_for_scenario(current_scenario, "a")

    inference_a.previous_key = colA.text_input(
        "Previous key A", inference_a.previous_key
    )
    colA.write(inference_a.__dict__)
    colA.dataframe(perform_inference(inference_a), height=500)

    colB.write("##### Input B")
    inference_b = get_inference_input_for_scenario(current_scenario, "b")

    inference_b.previous_key = colB.text_input(
        "Previous key B", inference_b.previous_key
    )
    colB.write(inference_b.__dict__)
    colB.dataframe(perform_inference(inference_b), height=500)



from __future__ import annotations

from typing import Optional

from pydantic import BaseModel


class RunPerformed(BaseModel):
    """
    Main event of the application.
    Identifies a search being executed
    """

    # name of the entry matched
    key: str
    # for when a query was typed by the user
    # @todo rename to something more meaningful
    query_input: str
    # for when it is started from a shortcut
    shortcut: str
    # unix timestamp
    timestamp: Optional[str] = None
    rank_uuid: Optional[str] = None
    rank_position: Optional[int] = None

    @staticmethod
    def get_schema():
        return "key string, query_input string, shortcut string, rank_uuid string"



#!/usr/bin/env python3

"""
This is a initial version of python_search _entries to give you a sense of what is possible.
Feel free to delete all the _entries here. You will want to versionate your _entries
"""
import datetime
import os

entries = {
    # NEW_ENTRIES_HERE
    "register new entry": {
        # cmd scripts will be executed in a terminal in the background
        "cmd": "python_search register_new from_clipboard",
    },
    # cli cmds will additional open a new terminal window and execute the  command
    "edit current project _entries source code": {
        "cli_cmd": f"python_search edit_main",
    },
    # urls will be opened in the browser
    "documentation about entities python search": {
        "url": "https://github.com/jeanCarloMachado/PythonSearch/blob/main/docs/entities.md"
    },
    "python search cli source code online reference": {
        "url": "https://shorturl.at/bdeK6"
    },
    # snippets copy  the content of the snippet to the clipboard when executed
    "date current today now copy": {
        "snippet": datetime.datetime.now().strftime("%Y-%m-%d %H:%M"),
        "i3_shortcut": "Control+Shift+0",
    },
    # file will open the file with the standard file handler for the file type in the system
    "edit current selected project _configuration of python search": {
        # you can have multiple projects in paralel with python search
        "file": f"{os.environ['HOME']}/.config/python_search/current_project",
    },
    "get python help function on object": {
        # pythons search is shipped with a multiplataform UI
        # that collects_input to questions and returns the string answered
        "cli_cmd": """python3 -c 'help($(collect_input giveTheObjectName))' """
    },
    # _entries are python code, you can import them from other python scripts
    # or you can generate them dynamically like in the example below
    # here we are generating different _entries for different environments (production, testing, development)
    "help python search manual": {"cli_cmd": "python_search --help"},
    **{
        f"get pods for {env}": {"cli_cmd": f"kubectl --context {env} get pods"}
        for env in ["production", "testing", "development"]
    },
}


from python_search.config import PythonSearchConfiguration

config = PythonSearchConfiguration(entries=entries)



from python_search.config import ConfigurationLoader
from python_search.data_ui.entries_page import extract_value_from_entry


class DataExporter:
    blacklisted_terms = [
        "http://",
        "https://",
        "email",
        "no key ",
        "insurance",
        "token",
        "private",
        "endereco",
        "password",
        "secret",
        "passport",
        "passaporte",
        "telefone",
        "phone",
        "celular",
        "cellphone",
        "cpf",
    ]

    def export_as_text(self):

        entries = ConfigurationLoader().load_entries()

        data = ""
        for key, value in entries.items():
            serializeable_value = extract_value_from_entry(value).replace("\n", " ")
            if any(
                [
                    blackelisted_entry in key.lower()
                    for blackelisted_entry in self.blacklisted_terms
                ]
            ):
                print("Skipping blacklisted key: ", key)
                continue

            if any(
                [
                    blackelisted_entry in serializeable_value.lower()
                    for blackelisted_entry in self.blacklisted_terms
                ]
            ):
                print("Skipping blacklisted value: ", serializeable_value)
                continue

            if "private" in value:
                continue

            data = data + f"{key}={serializeable_value}\n"

        with open("exported_entries.txt", "w") as f:
            f.write(data)


if __name__ == "__main__":
    import fire

    fire.Fire()



from __future__ import annotations

import subprocess

import streamlit as st

from python_search.data_ui.app_functions import restart_app


def extract_value_from_entry(entry):
    result = ""
    if "url" in entry:
        result = str(entry["url"])
    if "snippet" in entry:
        result = str(entry["snippet"])
    if "file" in entry:
        result = str(entry["file"])
    if "callable" in entry:
        result = str(entry["callable"])
    if "cmd" in entry:
        result = str(entry["cmd"])
    if "cli_cmd" in entry:
        result = str(entry["cli_cmd"])
    return result


def load_homepage():
    from python_search.config import ConfigurationLoader

    entries = ConfigurationLoader().load_config().commands

    col1, col2, col3 = st.columns([1, 1, 1])

    with col1:
        if st.button("Sync current host"):
            result = subprocess.check_output(
                "/src/sync_hosts.py sync", shell=True, text=True
            )
            st.write(f"Result: {result}")
            restart_app()
    with col2:
        if st.button("Restart"):
            restart_app()

    with col3:
        if st.checkbox("Add New Entry"):
            open_add_new = True
        else:
            open_add_new = False

    if open_add_new:
        key = st.text_input("Key")
        value = st.text_input("Value")
        create = st.button("Create")
        if create:
            cmd = f"python_search register_new register --key='{key}' --value='{value}' --tag=DataApp_Entry"
            st.write("Running: ", cmd)
            result = subprocess.check_output(cmd, shell=True, text=True)
            st.write(f"Result: {result}")

    st.write(" ## Entries")
    search = st.text_input("Search").lower()

    selected_tags = st.multiselect(
        "Tags", ConfigurationLoader().load_config().get_default_tags()
    )
    has_filter = len(search) > 0 or len(selected_tags) > 0
    limit = 50 if not has_filter else None

    rendered = 0
    for key, value in entries.items():
        if limit and rendered > limit:
            break

        value_str = extract_value_from_entry(value)
        tags = " ".join(value.get("tags", []))

        if not set(selected_tags).issubset(set(value.get("tags", []))):
            continue

        if (
            search
            and (search not in key)
            and search not in value_str
            and search not in tags
        ):
            continue

        col_key, col_value, col_tags = st.columns((1, 2, 1))
        col_key.write(key)
        col_value.write(value_str)
        col_tags.write(tags)

        rendered += 1
    st.write(f"**Total entries displayed**: {rendered}")



import os
import time


class Mac:
    """
    Mac support for tool iCanHazShortcut

    """

    START_SHORTCUT_NUMBER = 1

    def __init__(self, configuration):
        self.configuration = configuration
        import os

        self.config_folder = f'{os.environ["HOME"]}/.config/iCanHazShortcut/'

    def generate(self, restart_app=True):
        print("Generating macos shortctus")

        if restart_app:
            print("Killing shortcut app")
            os.system("pkill -f iCanHa")
            time.sleep(2)

        shortcut_found = False
        # starts with number 2 as number 1 is static in config.part1

        import shutil

        shutil.copyfile(
            f"{self.config_folder}/config.ini.part1", f"{self.config_folder}/config.ini"
        )

        content_to_write = ""
        for key, content in list(self.configuration.commands.items()):
            if not type(content) is dict:
                continue

            if "mac_shortcut" in content:
                shortcut_found = True
                content_to_write += self._add_shortcut(
                    content["mac_shortcut"], key, Mac.START_SHORTCUT_NUMBER
                )
                Mac.START_SHORTCUT_NUMBER += 1

            if "mac_shortcuts" in content:
                shortcut_found = True
                for shortcut in content["mac_shortcuts"]:
                    content_to_write += self._add_shortcut(
                        shortcut, key, Mac.START_SHORTCUT_NUMBER
                    )
                    Mac.START_SHORTCUT_NUMBER += 1

        if not shortcut_found:
            print("No shortcuts found for mac")
            return
        else:
            config_file = f"{self.config_folder}config.ini"
            with open(config_file, "a") as myfile:
                print(config_file, content_to_write)
                myfile.write(content_to_write)

        os.system(f"cd {self.config_folder} ; add_bom_to_file.sh config.ini")
        time.sleep(1)

        # restart shortcut app
        if restart_app:
            print("Restarting shortcut app")
            os.system("open -a iCanHazShortcut")

        print(f"Done! {Mac.START_SHORTCUT_NUMBER}  shortcuts generated")

        import subprocess

        number_of_items = subprocess.getoutput(
            f"grep '\[shortcut' {self.config_folder}config.ini | wc -l"
        )
        print(f"In the file we found {number_of_items} items")

    def _add_shortcut(self, shortcut: str, key: str, shortcut_number) -> str:
        print(f"Generating shortcut for {key}")
        return self._entry(shortcut, key, shortcut_number)

    def _entry(self, shortcut, key, number) -> str:
        shortcut = shortcut
        return f"""

[shortcut{number}]
shortcut = {shortcut}
action = {key}{number}
command =  /Users/jean.machado@getyourguide.com/projects/PythonSearch/log_command.sh python_search run_key "{key}"
workdir =
enabled = yes
"""



""" Module responsible for the logic of editing entry files """
import logging
import subprocess
from typing import Optional

from python_search.apps.terminal import Terminal
from python_search.interpreter.cmd import CmdInterpreter


class EditKey:
    """
    Set of commands to edit the _entries
    """

    def __init__(self, configuration):
        self.configuration = configuration

    def edit_key(self, key, dry_run=False):
        """
        Edits the _configuration files by searching the text
        """
        print(f"Editing key {key}")
        if not key:
            self.edit_default()
            return

        key = key.split(":")

        if not len(key):
            self.edit_default()
            return

        key = key[0]
        # needs to be case insensitive search
        cmd = f"ack -i '{key}' {self.configuration.get_project_root()} --py || true"
        logging.info(f"Command: {cmd}")
        result_shell = subprocess.check_output(cmd, shell=True, text=True)

        if not result_shell:
            logging.info("Could not find match edit main file")
            self._edit_config(self.configuration.get_source_file(), dry_run)
            return

        file, line, *_ = result_shell.split(":")
        print(f"Editng file and line {file}, {line}")

        self._edit_config(file, line)

    def search_entries_directory(self, key=None):
        entry = {
            "cli_cmd": f"fzf_directory.sh {self.configuration.get_project_root()}",
            "directory": self.configuration.get_project_root(),
        }

        CmdInterpreter(entry).interpret_default()

    def edit_default(self):
        self._edit_config(self.configuration.get_project_root() + "/entries_main.py")

    def _edit_config(self, file_name: str, line: Optional[int] = 30, dry_run=False):
        """
        edit a configuration file given the name and line
        """

        # @ todo make this editor generic

        cmd: str = (
            f"MY_TITLE='GrimorieSearchRun' kitty {Terminal.GENERIC_TERMINAL_PARAMS} bash -c 'cd"
            f" {self.configuration.get_project_root()} "
            f"; {self._get_open_text_editor_command(file_name, line)} "
        )
        print(cmd)

        if dry_run:
            logging.info(f"Command to edit file: {cmd}")
            return

        import os

        os.system(cmd)

    def _get_open_text_editor_command(self, file, line):
        if self.configuration.get_text_editor() == 'vim':
            return f"{self.configuration.get_text_editor()} {file} +{line}'"
        elif self.configuration.get_text_editor() == 'docker_nvim':
            return f"{self.configuration.get_text_editor()} {file} --line={line}'"
        else:
            # if is not a known editor just open the file
            return f"{self.configuration.get_text_editor()} {file}"



import logging
import re

from python_search.context import Context
from python_search.exceptions import CommandDoNotMatchException
from python_search.interpreter.base import BaseInterpreter
from python_search.interpreter.cmd import CmdInterpreter
from python_search.interpreter.file import FileInterpreter
from python_search.interpreter.group import GroupInterpreter
from python_search.interpreter.python import PythonInterpreter
from python_search.interpreter.snippet import SnippetInterpreter
from python_search.interpreter.url import UrlInterpreter

INTERPRETERS_IN_ORDER = [
    UrlInterpreter,
    FileInterpreter,
    GroupInterpreter,
    SnippetInterpreter,
    PythonInterpreter,
    CmdInterpreter,
]


class InterpreterMatcher:
    """
    Matches a query with an entry interpreter
    """

    _instance = None

    @staticmethod
    def build_instance(configuration):
        """
        Singleton. Initializes a _configuration a context and a interpreter

        :return:
        """
        if not InterpreterMatcher._instance:
            context = Context.get_instance()
            InterpreterMatcher._instance = InterpreterMatcher(configuration, context)

        return InterpreterMatcher._instance

    def __init__(self, configuration, context: Context):
        self._configuration = configuration
        self.context = context
        self.context.set_interpreter(self)
        self._interpreters = INTERPRETERS_IN_ORDER

    def get_interpreter(self, given_input: str) -> BaseInterpreter:
        """
        Given the string content, returns the best matched interpreter.
        Returns the instance of the matched interpreter given an text input
        """
        self.context.set_input(given_input)

        try:
            # tries to get the real key if it exists
            key = self._get_key(given_input)
            given_input = self._configuration.get_command(key)
        except Exception as e:
            logging.error(e)

        return self._match_interpreter(given_input)

    def get_interpreter_from_type(self, type: str):
        """
        From a type given in the ui via string returns the matching interpreter type

        """

        if type == "Snippet":
            return SnippetInterpreter

        if type == "Url":
            return UrlInterpreter

        if type == "Cmd":
            return CmdInterpreter

        if type == "File":
            return FileInterpreter

        raise Exception(f"Could not find a matching interpreter for string {type}")

    def default(self, given_input: str):
        """
        Applies the default behaviour to an interpreter
        """

        specific_interpreter = self.get_interpreter(given_input)

        return specific_interpreter.default()

    def clipboard(self, given_input: str):
        specific_interpreter: BaseInterpreter = self.get_interpreter(given_input)
        return specific_interpreter.interpret_clipboard()

    def _match_interpreter(self, cmd) -> BaseInterpreter:
        for interpreter in self._interpreters:
            try:
                logging.info(f"Trying to construct {interpreter}")
                command_instance = interpreter(cmd, self.context)
                logging.info(f"Matched command instance {command_instance}")
                return command_instance
            except CommandDoNotMatchException as e:
                pass

        raise Exception("Received a dict but did not match any type")

    def _get_key(self, given_input) -> str:
        """
        @todo have a global way to match the keys and move the logic out of here
        this already caused a bug on key contents not being able to be copied
        """
        key_value = re.compile("([A-Za-z0-9 _-]+):(.*)")
        matches_kv = key_value.search(given_input)
        key = given_input
        if matches_kv:
            key = matches_kv.group(1)
        return key



from __future__ import annotations

import streamlit as st


from python_search.data_ui.app_functions import check_password
from python_search.events.run_performed.dataset import EntryExecutedDataset

from python_search.data_ui.training_page import load_training_page

open_page = "home"
def init():
    st.set_page_config(initial_sidebar_state="collapsed", layout="wide")

def sidebar():
    global open_page
    with st.sidebar:

        if st.button("HomePage"):
            open_page = "home"

        if st.button("Results evaluation"):
            open_page = "results"

        if st.button("Training Dataset"):
            open_page = "training"

        if st.button("Searches Performed Dataset"):
            open_page = "searches_performed_dataset"


def render_page():
    global open_page
    if open_page == "home":
        from python_search.data_ui.entries_page import load_homepage

        load_homepage()

    if open_page == "training":
        load_training_page()

    if open_page == "results":
        import python_search.data_ui.results_page as results_page

        results_page.load_results_page()

    if open_page == "searches_performed_dataset":
        st.write("## Searches performed dataset")
        search_performed_df = EntryExecutedDataset().load_clean()
        pdf = search_performed_df.toPandas()
        st.dataframe(pdf)


if __name__ == '__main__':
    init()
    sidebar()
    if check_password():
        render_page()



from __future__ import annotations
import json
import uuid
from typing import Optional, Tuple

import numpy as np
from pydantic import BaseModel

from python_search.entry_type.entity import EntryType
from python_search.infrastructure.arize import Arize

PredictionUuid = str


class PredictEntryTypeInference:
    PRODUCTION_RUN_ID = "004224c854464ec296b5f648bd3f74f5"

    def __init__(self):
        # Instantiate an Arize Client object using your API and Space keys
        self._arize_client = Arize().get_client()

    def predict_entry_type_from_dict(self, entry: dict) -> EntryType:
        return self.predict_entry_type(EntryData(**entry))

    def predict_entry_type(
        self, entry_data: EntryData
    ) -> Tuple[EntryType, PredictionUuid]:

        from python_search.search.models import PythonSearchMLFlow

        model = PythonSearchMLFlow().get_entry_type_classifier(
            run_id=PredictEntryTypeInference.PRODUCTION_RUN_ID
        )
        from python_search.search.next_item_predictor.features.entry_embeddings.entry_embeddings import (
            create_embeddings_from_strings,
        )

        data = create_embeddings_from_strings([entry_data.content])

        from python_search.entry_type.entry_type_pipeline import Pipeline

        X = np.zeros([1, Pipeline.INPUT_DIMENSIONS])

        has_pipe = "|" in entry_data.content
        has_double_minus = "--" in entry_data.content

        X[0] = np.concatenate(
            (
                data[0],
                np.asarray([1 if has_pipe else 0]),
                np.asarray([1 if has_double_minus else 0]),
            )
        )

        result = model.predict(X)
        value, prediction_label = get_value_and_label(result[0])

        prediction_uuid = str(uuid.uuid4())
        if Arize.is_installed():
            from arize.utils.types import ModelTypes, Environments, Embedding

            arize_result = self._arize_client.log(
                model_id=Arize.MODEL_ID,
                model_version=Arize.MODEL_VERSION,
                model_type=ModelTypes.SCORE_CATEGORICAL,
                environment=Environments.PRODUCTION,
                prediction_id=prediction_uuid,
                features={"has_pipe": has_pipe, "has_double_minus": has_double_minus},
                embedding_features={
                    "content": Embedding(vector=data[0], data=entry_data.content)
                },
                prediction_label=(prediction_label, float(value)),
            )
            Arize.arize_responses_helper(arize_result)

        print("Predicted label: ", prediction_label)

        return prediction_label, prediction_uuid


class ClassifierInferenceClient:
    def predict_from_content(self, content: str):
        return self.predict(EntryData(content=content))

    def predict(self, data: EntryData) -> Optional[Tuple[EntryType, PredictionUuid]]:
        import requests

        try:
            result = requests.post(
                url="http://localhost:8000/entry_type/classify", json=data.__dict__
            )
            data = json.loads(result.text)
            return data["predicted_type"], data["prediction_uuid"]
        except BaseException as e:
            print(f"Logging results failed, reason: {e}")
            return None


def get_value_and_label(prediction_result) -> Tuple[float, EntryType]:
    max_val = max(prediction_result)

    for k, v in enumerate(prediction_result):
        if max_val == v:
            result = EntryType.from_categorical(k)
            break
    return max_val, result.value


class EntryData(BaseModel):
    content: str
    key: Optional[str]


def main():
    import fire

    fire.Fire()


if __name__ == "__main__":
    main()



import json


def decode_serialized_data_from_entry_text(entry_text: str, logger=None) -> dict:
    key = entry_text.split(":")[0]
    serialized_content = entry_text[len(key) + 1 :]
    try:
        result = json.loads(serialized_content)
        logger.info(f"Decoded serialized content worked {result}")

        return result
    except Exception as e:
        message = f"Failed with error {e} wile decoding the followingthe entry_text '{entry_text}' "
        if logger:
            logger.info(message)

        return {}



import os


class WindowManager:
    """
    Abstract specific window manager calls and initiallization
    """

    @staticmethod
    def load_from_environment() -> "WindowManager":
        if WindowManager.is_i3():
            return I3()

        return Gnome()

    @staticmethod
    def is_i3():
        return not WindowManager.is_gnome()

    @staticmethod
    def is_gnome():
        return 0 == os.system("wmctrl -m | grep -i gnome ")

    def hide_window(self, title):
        raise Exception("Not implemented")


class I3(WindowManager):
    """
    Contains implementation of window manager necessary functions for i3
    The same interfaces could be implemented for other window managers
    """

    def focus_on_window_with_title(self, title) -> bool:
        """
        Try to focus on the window with the passed title, returns True if successful
        """
        cmd = f'wmctrl -a "{title}" '
        print("Focus on window with cmd:", cmd)

        result = 0 == os.system(cmd)

        if result:
            self.show_window(title)

        return result

    def show_window(self, title) -> bool:
        return 0 == os.system(
            f"unset I3SOCK ; i3-msg '[title=\"{title}\"] scratchpad show'"
        )

    def hide_window(self, title) -> bool:
        return 0 == os.system(
            f"sleep 0.1; i3-msg '[title=\"{title}\"]  move scratchpad'"
        )


class Gnome:
    """
    Contains implementation of window manager necessary functions for gnome
    The same interfaces could be implemented for other window managers
    """

    def hide_window(self, title) -> bool:
        return 0 == os.system(f"xdotool search --name '{title}' windowminimize")



#!/usr/bin/env python3
import os


def build():
    result = os.system("docker build . -t ps:latest ")
    if result == 0:
        print("Build successful")
    else:
        raise Exception("Build failed")


def build_and_run():
    build()
    run()


def run(cmd="", entrypoint="", port="", restart=False, extra_env_vars=None):

    if entrypoint:
        entrypoint = f" --entrypoint '{entrypoint}'"

    restart_exp = ""
    if restart:
        print("Adding restart flag")
        restart_exp = " --restart always "

    if port:
        port = f" -p {port}"

    volumes = " ".join(
        [
            " -v $HOME/projects/PythonSearch:/src ",
            " -v $HOME/.ssh:/root/.ssh ",
            " -v $HOME/projects/PySearchEntries/:/entries ",
            " -v $HOME/.PythonSearch:/root/.PythonSearch ",
            " -v $HOME/.ddataflow:/root/.ddataflow",
            " -v $HOME/.PythonSearch/container_cache/:/root/.cache ",
            " -v $HOME/.data:/root/.data" " -v $HOME/.gitconfig:/root/.gitconfig",
        ]
    )

    env_vars = [
        " -e 'PS_ENTRIES_HOME=/entries' ",
        " -e ARIZE_API_KEY=$ARIZE_API_KEY ",
        " -e ARIZE_SPACE_KEY=$ARIZE_SPACE_KEY ",
        " -e PS_WEB_PASSWORD=$PS_WEB_PASSWORD",
    ]

    env_vars = env_vars + extra_env_vars if extra_env_vars else env_vars

    environment_variables = " ".join(env_vars)

    LIMIT_CPU = 8
    cmd = f"docker run {port} {restart_exp} --expose=8000 --expose 4040 --expose 6380 --cpus={LIMIT_CPU} {environment_variables} -it {volumes} {entrypoint} ps {cmd}"
    print("Cmd: " + cmd)
    os.system(cmd)


def sh():
    shell()


def shell():
    run(entrypoint="/bin/bash")


def run_jupyter(with_token=False, restart=False):

    token_expression = " --NotebookApp.token=''"
    if with_token:
        token_expression = ""
    run(
        cmd=f"jupyter lab --allow-root --ip '*' --notebook-dir / {token_expression} --NotebookApp.password=''",
        port="8888:8888",
        restart=restart,
    )


def run_mlflow(restart=False):
    run(
        cmd="mlflow ui --backend-store-uri file:/entries/mlflow --port 5001 --host '0.0.0.0' ",
        port="5001:5001",
        restart=restart,
    )


def run_webserver(restart=False, force_restart=False):
    if force_restart:
        print("Stopping previously running container")
        os.system("docker stop $(docker ps | grep -i 8000 | cut -d ' ' -f1) ")
    run(
        cmd="python_search_webapi",
        port="8000:8000",
        restart=restart,
    )


def run_streamlit(restart=False, disable_password=False):
    run(
        cmd="streamlit run python_search/data_ui/main.py --server.address=0.0.0.0  --server.port=8501",
        port="8501:8501",
        restart=restart,
        extra_env_vars=[" -e 'PS_DISABLE_PASSWORD=1' "] if disable_password else None,
    )

def run_calendar(restart=True, disable_password=False):
    run(
        cmd="streamlit run /entries/calendar_app.py --server.address=0.0.0.0  --server.port=8502",
        port="8502:8502",
        restart=restart,
        extra_env_vars=[" -e 'PS_DISABLE_PASSWORD=1' "] if disable_password else None,
    )

def start():
    import fire

    fire.Fire()


if __name__ == "__main__":
    start



""" Centralize all exceptions definitions to help in the _entries discovery """


class MenuException(Exception):
    @staticmethod
    def given_empty_value():
        return MenuException("No option selected in search ui!")


class CommandDoNotMatchException(Exception):
    @staticmethod
    def not_valid_command(entity, cmd):
        return CommandDoNotMatchException(
            f"Not Valid {entity.__class__.__name__} command {cmd}"
        )


class RunException(Exception):
    @staticmethod
    def key_does_not_exist(key: str):
        return RunException(f"Key does not exist: {key}")


class RegisterNewException(Exception):
    @staticmethod
    def empty_content():
        return RegisterNewException(f"Will not register as content looks too small")


class MissingConfigException:
    @staticmethod
    def configuration_not_set():
        raise MissingConfigException(
            "The python search configuration was not found. Run python_search setup to initialize a new config"
        )


from functools import wraps


def notify_exception():
    """
    Will let you know when a function is called or returned error as a desktop notification
    """

    def _(func):
        @wraps(func)
        def __(*args, **kwargs):
            try:
                result = func(*args, **kwargs)
            except Exception as e:
                from python_search.apps.notification_ui import send_notification

                send_notification(f"Exception: {str(e)}")
                raise e
            return result

        return __

    return _



#!/usr/bin/env python3

from typing import List


class RecentKeys:
    """
    Contains the latest used keys and the API to add new ones.
    """

    _blacklisted_items = ["python search main entry"]

    _used_keys = []

    def get_latest_used_keys(self) -> List[str]:
        """
        return a list of unike used keys ordered by the last time they were used
        the most recent in the top.
        """
        return RecentKeys._used_keys

    @staticmethod
    def add_latest_used(key):
        """adds to the list"""
        if key in RecentKeys._blacklisted_items:
            return

        if key in RecentKeys._used_keys:
            RecentKeys._used_keys.remove(key)

        RecentKeys._used_keys = [key] + RecentKeys._used_keys


if __name__ == "__main__":
    import fire

    fire.Fire(RecentKeys)



from __future__ import annotations

import re
from typing import List

from python_search.apps.notification_ui import send_notification
from python_search.config import PythonSearchConfiguration
from python_search.context import Context
from python_search.events.run_performed import RunPerformed
from python_search.events.run_performed.writer import LogRunPerformedClient
from python_search.interpreter.cmd import CmdInterpreter
from python_search.interpreter.interpreter_matcher import InterpreterMatcher
from python_search.logger import setup_run_key_logger, StreamToLogger
from python_search.exceptions import notify_exception
from python_search.search_ui.serialized_entry import (
    decode_serialized_data_from_entry_text,
)

logger = setup_run_key_logger()


class EntryRunner:
    """
    Responsible to execute the _entries matched
    """

    def __init__(self, configuration: PythonSearchConfiguration):
        self.configuration = configuration

    @notify_exception()
    def run(
        self,
        entry_text: str,
        query_used: str = "",
        force_gui_mode=False,
        gui_mode=False,
        from_shortcut=False,
    ):
        """
        Runs an entry given its name or its partial name.

        Parameters:
            key: As it comes from FZF they is a str pair of key_name : {metadata}
            entry_rank_position: accounts for where the entry was when it was executed, if passed it will be used for
            from_shortcut means that the key execution was triggered by a desktop shortcut
        """
        key = entry_text.split(":")[0] if ":" in entry_text else entry_text

        logger.info("Arrived at run key")
        # if there are : in the line just take all before it as it is
        # usually the key from fzf, and our keys do not accept :

        if from_shortcut:
            send_notification(f"{key}")

        metadata = decode_serialized_data_from_entry_text(entry_text, logger)
        logger.info(f"Decoded metadata {metadata}")
        rank_position = metadata.get("position")

        # when there are no matches we actually will use the query and interpret it
        if not key and query_used:
            CmdInterpreter({"cli_cmd": query_used}).interpret_default()
            return

        matches = self._matching_keys(key)

        if not matches:
            raise Exception(f"No key matches you given requested key: {key}")

        logger.info(
            f"""
            Matches of key: {key}
            matches: {matches}
            Query used: {query_used}
            Rank Position: {rank_position}
        """
        )

        if force_gui_mode or gui_mode:
            Context.get_instance().enable_gui_mode()

        real_key: str = matches[0]

        if len(matches) > 1:
            real_key = min(matches, key=len)
            send_notification(
                f"Multiple matches for this key {matches} using the smaller"
            )

        result = InterpreterMatcher.build_instance(self.configuration).default(real_key)

        logger.info("Passed interpreter")
        run_performed = RunPerformed(
            key=key,
            query_input=query_used,
            shortcut=from_shortcut,
            rank_uuid=metadata.get("uuid"),
            rank_position=metadata.get("position"),
        )
        logger.info(f"Run performed = {run_performed}")
        LogRunPerformedClient().send(run_performed)
        return result

    def _matching_keys(self, key: str) -> List[str]:
        """
        give a key it will give suggestions that matches
        """

        key = generate_identifier(key)
        key_regex = re.compile(key)

        matching_keys = []
        for registered_key in self.configuration.get_keys():
            encoded_registered_key = generate_identifier(registered_key)
            matches_kv_encoded = key_regex.search(encoded_registered_key)
            if matches_kv_encoded:
                logger.info(f"{key} matches {encoded_registered_key}")
                matching_keys.append(registered_key)

        return matching_keys


def generate_identifier(string):
    """
    strip the string from all special characters lefting only [A-B-09]
    """
    result = "".join(e for e in string if e.isalnum())
    result = result.lower()

    return result



import os
from typing import Literal, Optional

from mlflow.entities import RunInfo

from python_search.config import DataConfig

BASE_MLFLOW_LOCATON = "/entries/mlflow"

NEXT_ITEM_PREDICTOR_PROJECT_NUMBER = "2"
ENTRY_TYPE_CLASSIFIER_PROJECT_NUMBER = "4"
ENTRY_DESCRIPTION_GENERATOR_PROJECT_NUMBER = "5"


class PythonSearchMLFlow:
    """
    Accessor to MLflow API

    """

    def __init__(self):

        self.debug = os.getenv("PS_DEBUG", False)
        import mlflow

        self.mlflow_instance = mlflow
        self.mlflow_instance.set_tracking_uri(f"file:{DataConfig.MLFLOW_MODELS_PATH}")

    def get_latest_next_predictor_run(self) -> RunInfo:
        from mlflow.tracking import MlflowClient

        experiment_name = DataConfig.NEXT_ITEM_EXPERIMENT_NAME

        client: MlflowClient = MlflowClient()
        experiment = client.get_experiment_by_name(experiment_name)
        if self.debug:
            print(f"Experiment id: {experiment.experiment_id}")
        runs = client.list_run_infos(experiment_id=experiment.experiment_id)
        return runs[0]

    def get_next_predictor_model(self, run_id: Optional[str] = None):
        from typing import Literal

        model_type: Literal["xgboost", "keras"] = "xgboost"
        # model_type: Literal["xgboost", "keras"] = "keras"

        if not run_id:
            run_id = self.get_latest_next_predictor_run().run_id

        if self.debug:
            print(f"Loading run id: {run_id}")

        path = f"{BASE_MLFLOW_LOCATON}/{NEXT_ITEM_PREDICTOR_PROJECT_NUMBER}/{run_id}/artifacts/model"

        if model_type == "keras":
            model = self.mlflow_instance.keras.load_model(path)
        else:
            model = self.mlflow_instance.xgboost.load_model(path)

        return model

    def get_entry_type_classifier(self, run_id):
        return self.mlflow_instance.keras.load_model(
            f"{BASE_MLFLOW_LOCATON}/{ENTRY_TYPE_CLASSIFIER_PROJECT_NUMBER}/{run_id}/artifacts/model"
        )

    def get_entry_description_geneartor(self, run_id):
        return self.mlflow_instance.keras.load_model(
            f"{BASE_MLFLOW_LOCATON}/{ENTRY_DESCRIPTION_GENERATOR_PROJECT_NUMBER}/{run_id}/artifacts/model"
        )

    def get_entry_description_geneartor_dict(self, run_id):
        return self.mlflow_instance.artifacts.load_dict(
            f"{BASE_MLFLOW_LOCATON}/{ENTRY_DESCRIPTION_GENERATOR_PROJECT_NUMBER}/{run_id}/artifacts/chars"
        )



import os
import concurrent.futures as cf


class Arize:

    MODEL_ID = "entry_type_classifier_v2"
    MODEL_VERSION = "v5"

    @staticmethod
    def is_installed() -> bool:
        try:
            import arize

            return True
        except:
            return False

    def get_client(self):
        if not Arize.is_installed():
            return None

        API_KEY = os.environ["ARIZE_API_KEY"]
        SPACE_KEY = os.environ["ARIZE_SPACE_KEY"]

        if not API_KEY or not SPACE_KEY:
            print(
                "Arize cannot be created without both API_KEY and SPACE_KEY, values: "
                + API_KEY
                + " "
                + SPACE_KEY
            )
            return

        from arize.api import Client

        return Client(space_key=SPACE_KEY, api_key=API_KEY)

    def arize_responses_helper(responses):
        """
        responses: a list of responses from Arize
        returns: None
        """
        responses = [responses]
        for response in cf.as_completed(responses):
            res = response.result()
            print(res.status_code)
            if res.status_code != 200:
                raise ValueError(f"failed with code {res.status_code}, {res.text}")
            print("Arize succeeded with 200 status code")



import logging
import sys


def setup_inference_logger():
    logger = logging.getLogger("inference")
    logger.setLevel(logging.DEBUG)

    # fh = logging.FileHandler("/tmp/inference.txt")
    # fh.setLevel(logging.DEBUG)
    # logger.addHandler(fh)

    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    logger.addHandler(ch)

    return logger


def setup_preview_logger():
    """
    Only writes to file
    """
    logger = logging.getLogger("preview")
    fh = logging.FileHandler("/tmp/preview.txt")
    fh.setLevel(logging.INFO)
    logger.addHandler(fh)

    return logger


def setup_run_key_logger():
    logger = logging.getLogger("run-key")
    fh = logging.FileHandler("/tmp/run_key.txt")
    fh.setLevel(logging.DEBUG)
    logger.addHandler(fh)

    return logger


def setup_data_writter_logger(event_name):
    logger = logging.getLogger(f"data-writer_{event_name}")
    logger.setLevel(logging.DEBUG)
    fh = logging.FileHandler(f"/tmp/data-writer_event_{event_name}")
    fh.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    logger.addHandler(fh)
    logger.addHandler(ch)

    return logger


class StreamToLogger(object):
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """

    def __init__(self, logger, level):
        self.logger = logger
        self.level = level
        self.linebuf = ""

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.level, line.rstrip())

    def flush(self):
        pass



from datetime import datetime

from colorama import Fore
from dateutil import parser

from python_search.config import ConfigurationLoader
from python_search.logger import setup_preview_logger
from python_search.search_ui.serialized_entry import (
    decode_serialized_data_from_entry_text,
)


class Preview:
    """
    Preview the entry in the search ui
    """

    def __init__(self):
        self.configuration = ConfigurationLoader()
        self.logger = setup_preview_logger()
        # do not send the errors to stderr, in the future we should send to kibana or a file

    def display(self, entry_text: str):
        """
        Prints the entry in the preview window

        """

        data = self._build_values_to_print(entry_text)
        self._print_values(data)

    def _print_values(self, data):
        print("")
        print(
            data["type"]
            + ": "
            + self._color_str(data["value"], self._get_color_for_type(data["type"]))
        )
        print(f"Key: {self._color_str(data['key'], Fore.YELLOW)}")
        if "description" in data:
            print(f"Description: {data['description']}")

        if "created_at" in data:
            print("Created at: " + data["created_at"])
            print("Entry Age: " + data["entry_age"])

        if "tags" in data:
            print(f"Tags: {data['tags']}")
        if "position" in data:
            print("Position: " + data["position"])

        if "uuid" in data:
            print("Uuid: " + data["uuid"][0:8])

    def _get_color_for_type(self, type):
        if type == "Cmd":
            return Fore.RED

        if type in ["Url", "File"]:
            return Fore.GREEN

        return Fore.BLUE

    def _build_values_to_print(self, entry_text) -> dict:
        key = entry_text.split(":")[0]
        # the entry content is after the key + a ":" character

        result = {}
        result["type"] = "Unknown"
        result["key"] = key
        entry_data = self._load_key_data(key)
        if "url" in entry_data:
            result["value"] = entry_data.get("url")
            result["type"] = "Url"

        if "file" in entry_data:
            result["value"] = entry_data.get("file")
            result["type"] = "File"

        if "snippet" in entry_data:
            result["value"] = entry_data.get("snippet")
            result["type"] = "Snippet"

        if "cli_cmd" in entry_data or "cmd" in entry_data:
            result["value"] = entry_data.get("cli_cmd", entry_data.get("cmd"))
            result["type"] = "Cmd"

        if "callable" in entry_data:
            value = entry_data.get("callable")
            import dill

            result["value"] = dill.source.getsource(value)

        if "created_at" in entry_data:

            creation_date = parser.parse(entry_data["created_at"])
            today = datetime.now()
            result["created_at"] = str(creation_date)
            result["entry_age"] = str(today - creation_date)

        if "description" in entry_data:
            result["description"] = entry_data["description"]

        decoded_content = decode_serialized_data_from_entry_text(
            entry_text, self.logger
        )

        if "position" in decoded_content:
            result["position"] = str(decoded_content["position"])

        result["tags"] = []
        if "tags" in entry_data:
            result["tags"] = " ".join(entry_data["tags"])

        if "tags" in decoded_content:
            result["tags"] = " ".join(decoded_content["tags"])

        if "uuid" in decoded_content:
            result["uuid"] = str(decoded_content["uuid"])

        return result

    def _color_str(self, a_string, a_color) -> str:
        return f"{a_color}{a_string}{Fore.RESET}"

    def _load_key_data(self, key):
        entries = self.configuration.load_entries()
        if not key in entries:
            print()
            print(
                (
                    f'Key "{self._color_str(key, Fore.RED)}" not found in python search data'
                )
            )
            import sys

            sys.exit(0)

        return entries[key]



import os
from typing import Any

from python_search.context import Context
from python_search.environment import is_mac
from python_search.exceptions import CommandDoNotMatchException
from python_search.interpreter.base import BaseInterpreter
from python_search.interpreter.cmd import CmdInterpreter


class FileInterpreter(BaseInterpreter):
    def __init__(self, cmd: Any, context: Context = None):
        self.context = context
        self.cmd = {}

        if type(cmd) == str:
            self.cmd = {"file": cmd}
            return

        if type(cmd) is dict and "file" in cmd:
            self.cmd = cmd
            return

        if type(cmd) is str and FileInterpreter.file_exists(cmd):
            self.cmd["file"] = cmd
            return

        raise CommandDoNotMatchException(
            f"Not Valid {self.__class__.__name__} command {cmd}"
        )

    def get_executable(self):
        filename, file_extension = os.path.splitext(self.cmd["file"])


        if file_extension in [".py", '.vim']:
            return 'docker_nvim'

        if is_mac():
            return "open"

        if os.path.isdir(self.cmd["file"]):
            return "nautilus"

        if file_extension == ".pdf":
            return "zathura"

        elif file_extension == ".ipynb":
            return "pycharm"

        return "docker_nvim"

    def interpret_default(self):
        executable = self.get_executable()

        cmd = f'{executable} "{self.cmd["file"]}"'

        final_cmd = self.cmd
        if executable in ["vim", 'docker_nvim']:
            final_cmd["cli_cmd"] = cmd
        else:
            final_cmd["cmd"] = cmd

        return CmdInterpreter(final_cmd, self.context).interpret_default()

    def copiable_part(self):
        return self.cmd["file"]

    @staticmethod
    def file_exists(candidate: str):
        return os.path.isfile(candidate) or os.path.isdir(candidate)



from __future__ import annotations
from typing import List, Union
from python_search.config import DataConfig
from python_search.data_collector import GenericDataCollector

import uuid

import datetime
from pydantic import BaseModel

EVENT_FOLDER = "ranking_generated"


NotSetYet = None


class RankingGenerated(BaseModel):
    # name of the entry matched
    ranking: List[str]
    # unix timestamp
    first: Union[str, NotSetYet] = NotSetYet
    uuid: Union[str, NotSetYet] = NotSetYet
    timestamp: Union[str, NotSetYet] = NotSetYet

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.uuid = str(uuid.uuid4())
        self.first = self.ranking[0]
        self.timestamp = str(datetime.datetime.now(datetime.timezone.utc).timestamp())


class RankingGeneratedWriter:
    def write(self, event: RankingGenerated):
        return GenericDataCollector().write(
            data=event.__dict__, table_name=EVENT_FOLDER
        )


class RankingGeneratedDataset:
    """
    Poit of acess for the searches performed
    This is the place to keep the source of truth for the schema as well
    """

    DATA_FOLDER = DataConfig.BASE_DATA_COLLECTOR_FOLDER + "/" + EVENT_FOLDER

    def __init__(self, spark=None):
        from pyspark.sql.session import SparkSession

        self.spark = spark if spark else SparkSession.builder.getOrCreate()

    def load(self):
        return self.spark.read.json(self.DATA_FOLDER)



import time


from python_search.apps.notification_ui import send_notification
from python_search.config import ConfigurationLoader, PythonSearchConfiguration


def get_random():
    config: PythonSearchConfiguration = ConfigurationLoader().load_config()
    print("Running reminder job")

    all_entries = config.commands

    reminders = {}
    for entry, content in all_entries.items():
        if "tags" in content and "Reminder" in content["tags"]:
            reminders[entry] = content

    import random

    random_key = random.choice(list(reminders.keys()))

    run_key(random_key)


from python_search.feature_toggle import FeatureToggle


def run_key(key):
    import os

    os.system(f"python_search run_key '{key}'")


def run_daemon():
    import schedule

    FeatureToggle().enable("reminders")
    schedule.every(15).minutes.do(get_random)

    entries = ConfigurationLoader().load_entries()
    for key, entry in entries.items():
        if "schedules" in entry:
            print("Found a schedule for key: " + key)
            for entry_schedule in entry["schedules"]:
                entry_schedule(schedule).do(run_key, key)

    print("entering loop")
    while True:
        feature_toggle = FeatureToggle()
        if feature_toggle.is_enabled("reminders"):
            print("Running pending")
            schedule.run_pending()
        else:
            print("Feature is disabled")

        print("Sleeping")
        time.sleep(1)


def main():
    import fire

    fire.Fire()


if __name__ == "__main__":
    main()



import logging
import os
from typing import Dict, List, Tuple

import numpy as np
from pyspark.sql import DataFrame

from python_search.config import ConfigurationLoader
from python_search.search.next_item_predictor.features.entry_embeddings import (
    InferenceEmbeddingsLoader,
)
from python_search.search.next_item_predictor.features.entry_embeddings.entry_embeddings import (
    create_key_indexed_embedding,
)
from python_search.search.next_item_predictor.inference.input import ModelInput
from python_search.search.next_item_predictor.training_dataset import TrainingDataset


class ModelTransform:
    """
    Transform takes an input and make it ready for inference

    From training dataset to -> _model input
    And from inference dataset -> _model input
    """

    _EMBEDDINGS_ENTRIES = 3

    # 2 embeddings of 384 dimensions
    # + 1 is for the month number
    # + 1 for entry number
    # + 1 for global popularity of previous key
    # + 1 for global popularity of previous_previous key
    DIMENSIONS = _EMBEDDINGS_ENTRIES * 384 + 1 + 1 + 1 + 1

    def __init__(self):
        configuration = ConfigurationLoader().load_config()
        self._all_keys = configuration.commands.keys()
        self.inference_embeddings = InferenceEmbeddingsLoader(self._all_keys)

    def transform_single(self, inference_input: ModelInput, all_keys) -> np.ndarray:
        """
        Transform the inference input into something that can be inferred.
        This is an element wise search.
        """

        previous_key_embedding = self.inference_embeddings.get_embedding_from_key(
            inference_input.previous_key
        )
        previous_previous_key_embedding = (
            self.inference_embeddings.get_embedding_from_key(
                inference_input.previous_previous_key
            )
        )

        # create an inference array for all keys
        X = np.zeros([len(self._all_keys), ModelTransform.DIMENSIONS])
        for i, key in enumerate(all_keys):
            key_embedding = self.inference_embeddings.get_embedding_from_key(key)
            if key_embedding is None:
                logging.warning(f"No content for key ({key})")
                continue

            X[i] = np.concatenate(
                (
                    key_embedding,
                    previous_key_embedding,
                    previous_previous_key_embedding,
                    np.asarray([inference_input.month]),
                    np.asarray([inference_input.hour]),
                    np.asarray([inference_input.times_used_previous]),
                    np.asarray([inference_input.times_used_previous_previous]),
                )
            )

        return X

    def transform_collection(
        self, dataset: DataFrame, use_cache=True
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Transform the dataset into X and Y
        Returns a pair with X, Y
        """
        print("Number of rows in the dataset: ", dataset.count())
        print(f"Dimensions of dataset = {ModelTransform.DIMENSIONS}")

        if use_cache:
            if not os.path.exists("/tmp/X.npy") or not os.path.exists("/tmp/Y.npy"):
                raise Exception("Cache not found")
            print("Using transformed data from cache")
            X = np.load("/tmp/X.npy", allow_pickle=True)
            Y = np.load("/tmp/Y.npy", allow_pickle=True)

            return X, Y

        embeddings_keys = self._create_embeddings_training_dataset(dataset)
        # one extra for the row number
        X = np.zeros([dataset.count(), ModelTransform.DIMENSIONS + 1])
        Y = np.empty(dataset.count())

        print("X shape:", X.shape)

        # transform the spark dataframe into a python iterable
        collected_rows = dataset.select(*TrainingDataset.COLUMNS).collect()

        for i, row in enumerate(collected_rows):
            X[i] = np.concatenate(
                [
                    # adds entry number so we can index and select the right row afterwards
                    # it gets deleted before training
                    np.asarray([row.entry_number]),
                    embeddings_keys[row.key],
                    embeddings_keys[row.previous_key],
                    embeddings_keys[row.previous_previous_key],
                    np.asarray([row.month]),
                    np.asarray([row.hour]),
                    np.asarray([row.times_used_previous]),
                    np.asarray([row.times_used_previous_previous]),
                ]
            )

            Y[i] = row.label

        X = np.where(np.isnan(X), 0.5, X)
        Y = np.where(np.isnan(Y), 0.5, Y)

        np.save("/tmp/X.npy", X)
        np.save("/tmp/Y.npy", Y)

        return X, Y

    def _create_embeddings_training_dataset(
        self, dataset: TrainingDataset
    ) -> Dict[str, np.ndarray]:
        """
        create embeddings with all training keys and keep them in memory
        """
        print("Creating embeddings of training dataset")

        # add embeddings to the dataset
        all_keys = self._get_all_keys_from_dataset(dataset)

        return create_key_indexed_embedding(all_keys)

    def _get_all_keys_from_dataset(self, dataset: TrainingDataset) -> List[str]:
        collected_keys = dataset.select("key", "previous_key").collect()

        keys = []
        for collected_keys in collected_keys:
            keys.append(collected_keys.key)
            keys.append(collected_keys.previous_key)

        return keys



class Shortcuts:
    """
    @todo move this out of the repo
    List of mapped i3 shortcuts, in gnome the syntax is different.
    """

    CTRL_W = "Control+w"
    CTRL_V = "Control+v"
    CTRL_SPACE = "Control+space"
    SUPER_SPACE = "Mod4+space"
    SUPER_A = "Mod4+A"
    SUPER_W = "Mod4+w"
    SUPER_D = "Mod4+d"
    SUPER_P = "Mod4+p"

    ALT_SPACE = "Mod1+space"
    ALT_SHIFT_SPACE = "Mod1+Shift+space"
    ALT_SHIFT_D = "Mod1+Shift+d"
    ALT_SHIFT_S = "Mod1+Shift+s"
    ALT_SHIFT_J = "Mod1+Shift+j"
    ALT_SHIFT_L = "Mod1+Shift+l"
    SUPER_H = "Mod4+h"
    SUPER_SHIFT_H = "Mod4+Shift+h"
    CTRL_SHIFT_H = "Control+Shift+h"
    CTRL_H = "Control+h"
    ALT_H = "Mod1+h"
    ALT_A = "Mod1+a"
    ALT_V = "Mod1+v"
    ALT_SHIFT_A = "Mod1+Shift+a"
    ALT_SHIFT_H = "Mod1+Shift+h"
    CTRL_B = "Control+b"
    CTRL_E = "Control+e"
    CTRL_P = "Control+p"
    CTRL_I = "Control+i"
    CTRL_K = "Control+k"
    CTRL_D = "Control+d"
    CTRL_L = "Control+l"
    CTRL_SHIFT_D = "Control+Shift+d"
    CTRL_SHIFT_K = "Control+Shift+k"
    CTRL_SHIFT_E = "Control+Shift+e"
    ALT_SHIFT_X = "Mod1+Shift+x"
    CTRL_5 = "Control+5"
    CTRL_Q = "Control+q"
    SUPER_T = "Mod4+t"
    SUPER_0 = "Mod4+0"
    SUPER_9 = "Mod4+9"
    SUPER_8 = "Mod4+8"
    SUPER_7 = "Mod4+7"

    SUPER_SHIFT_W = "Mod4+Shift+w"
    SUPER_SHIFT_V = "Mod4+Shift+v"

    CTRL_SHIFT_N = "Control+Shift+n"
    CTRL_SHIFT_9 = "Control+Shift+9"
    CTRL_SHIFT_X = "Control+Shift+x"
    CTRL_SHIFT_F = "Control+Shift+f"
    CTRL_SHIFT_8 = "Control+Shift+8"
    CTRL_SHIFT_7 = "Control+Shift+7"
    CTRL_SHIFT_6 = "Control+Shift+6"
    CTRL_SHIFT_5 = "Control+Shift+5"
    CTRL_SHIFT_4 = "Control+Shift+4"
    CTRL_SHIFT_3 = "Control+Shift+3"
    CTRL_SHIFT_1 = "Control+Shift+1"

    CTRL_SHIFT_G = "Control+Shift+g"
    CTRL_SHIFT_P = "Control+Shift+p"
    CTRL_SHIFT_W = "Control+Shift+w"
    CTRL_ALT_T = "Control+Mod1+t"
    ALT_SHIFT_T = "Mod1+Shift+t"

    ALT_SHIFT_I = "Mod1+Shift+i"
    ALT_SHIFT_0 = "Mod1+Shift+0"
    CTRL_SHIFT_0 = "Control+Shift+0"
    ALT_SHIFT_Z = "Mod1+Shift+z"
    ALT_K = "Mod1+k"
    ALT_O = "Mod1+o"
    ALT_T = "Mod1+t"
    ALT_SHIFT_K = "Mod1+Shift+k"
    ALT_D = "Mod1+d"
    ALT_I = "Mod1+i"
    ALT_P = "Mod1+p"
    ALT_SHIFT_P = "Mod1+Shift+p"
    ALT_W = "Mod1+w"
    ALT_SHIFT_W = "Mod1+Shift+w"
    ALT_U = "Mod1+u"
    ALT_N = "Mod1+n"
    ALT_M = "Mod1+m"
    SUPER_M = "Mod4+m"
    ALT_SHIFT_M = "Mod1+Shift+m"
    ALT_F = "Mod1+f"
    ALT_SHIFT_F = "Mod1+Shift+f"
    ALT_E = "Mod1+e"
    ALT_C = "Mod1+c"
    ALT_SHIFT_J = "Mod1+Shift+j"
    CTRL_J = "Control+j"
    CTRL_SHIFT_J = "Control+Shift+j"
    ALT_SHIFT_E = "Mod1+Shift+e"
    ALT_SHIFT_N = "Mod1+Shift+n"

    ALT_J = "Mod1+j"
    ALT_G = "Mod1+g"
    ALT_B = "Mod1+b"
    ALT_SHIFT_B = "Mod1+Shift+b"
    SUPER_B = "Mod4+b"
    SUPER_SHIFT_B = "Mod4+Shift+b"
    SUPER_SHIFT_1 = "Mod4+Shift+1"
    SUPER_SHIFT_2 = "Mod4+Shift+2"
    SUPER_SHIFT_3 = "Mod4+Shift+3"
    SUPER_SHIFT_4 = "Mod4+Shift+4"
    ALT_L = "Mod1+l"
    ALT_SHIFT_L = "Mod1+Shift+l"
    ALT_SHIFT_G = "Mod1+Shift+g"

    CTRL_1 = "Control+1"
    CTRL_2 = "Control+2"
    CTRL_3 = "Control+3"
    CTRL_4 = "Control+4"
    CTRL_5 = "Control+5"
    CTRL_6 = "Control+6"
    CTRL_7 = "Control+7"
    CTRL_8 = "Control+8"
    CTRL_9 = "Control+9"
    CTRL_0 = "Control+0"

    ALT_1 = "Mod1+1"
    ALT_2 = "Mod1+2"
    ALT_3 = "Mod1+3"
    ALT_4 = "Mod1+4"
    ALT_5 = "Mod1+5"
    ALT_6 = "Mod1+6"
    ALT_7 = "Mod1+7"
    ALT_8 = "Mod1+8"
    ALT_9 = "Mod1+9"
    ALT_0 = "Mod1+0"

    ALT_SHIFT_1 = "Mod1+Shift+1"
    ALT_SHIFT_2 = "Mod1+Shift+2"
    ALT_SHIFT_3 = "Mod1+Shift+3"
    ALT_SHIFT_4 = "Mod1+Shift+4"
    ALT_SHIFT_5 = "Mod1+Shift+5"
    ALT_SHIFT_6 = "Mod1+Shift+6"
    ALT_SHIFT_7 = "Mod1+Shift+7"
    ALT_SHIFT_8 = "Mod1+Shift+8"
    ALT_SHIFT_9 = "Mod1+Shift+9"
    ALT_SHIFT_0 = "Mod1+Shift+0"

    CTRL_1 = "Control+1"
    CTRL_2 = "Control+2"
    CTRL_3 = "Control+3"
    CTRL_4 = "Control+4"
    CTRL_5 = "Control+5"
    CTRL_6 = "Control+6"
    CTRL_7 = "Control+7"
    CTRL_8 = "Control+8"
    CTRL_9 = "Control+9"
    CTRL_0 = "Control+0"

    CTRL_SHIFT_1 = "Control+Shift+1"
    CTRL_SHIFT_2 = "Control+Shift+2"
    CTRL_SHIFT_3 = "Control+Shift+3"
    CTRL_SHIFT_4 = "Control+Shift+4"
    CTRL_SHIFT_5 = "Control+Shift+5"
    CTRL_SHIFT_6 = "Control+Shift+6"
    CTRL_SHIFT_7 = "Control+Shift+7"
    CTRL_SHIFT_8 = "Control+Shift+8"
    CTRL_SHIFT_9 = "Control+Shift+9"
    CTRL_SHIFT_0 = "Control+Shift+0"

    SUPER_U = "Mod4+u"
    SUPER_E = "Mod4+e"
    CAPS_E = "Mod3+e"
    SUPER_SHIFT_E = "Mod4+Shift+e"
    SUPER_SHIFT_C = "Mod4+Shift+c"
    SUPER_SHIFT_T = "Mod4+Shift+t"
    SUPER_G = "Mod4+g"
    SUPER_SHIFT_G = "Mod4+Shift+g"
    SUPER_Y = "Mod4+y"
    SUPER_I = "Mod4+i"
    SUPER_SHIFT_I = "Mod4+Shift+i"
    SUPER_SHIFT_O = "Mod4+Shift+o"
    SUPER_S = "Mod4+s"
    SUPER_C = "Mod4+c"

    SUPER_R = "Mod4+r"
    SUPER_SHIFT_R = "Mod4+Shift+r"

    ALT_R = "Mod1+r"
    ALT_SHIFT_R = "Mod1+Shift+r"
    ALT_SHIFT_V = "Mod1+Shift+v"

    @staticmethod
    @property
    def SUPER_L():
        raise Exception("Key mapped by system, should not be used")

    @staticmethod
    @property
    def SUPER_O():
        raise Exception("Key mapped by system, should not be used")

    @staticmethod
    @property
    def SUPER_V():
        raise Exception("Key mapped by system, should not be used")



#!/usr/bin/env python3
import os
from subprocess import Popen

from python_search.apps.notification_ui import send_notification


class Infrastructure:
    def __init__(self):
        self._RUNNING_INFO = {
            "web_api": {"cmd": "python_search_webapi", "log": "/tmp/log_webserver"},
            "reminders": {
                "cmd": "reminders run_daemon",
                "log": "/tmp/log_reminders",
            },
            "zookeeper": {
                "cmd": "zookeeper-server-start /opt/homebrew/etc/kafka/zookeeper.properties",
                "log": "/tmp/log_zoookeeper",
            },
            "kafka": {
                "cmd": "kafka-server-start /opt/homebrew/etc/kafka/server.properties",
                "log": "/tmp/log_kafka",
            },
            "redis": {
                "cmd": "cd /tmp ; redis-server /opt/homebrew/Cellar/redis/7.0.3/.bottle/etc/redis.conf",
                "log": "/tmp/log_redis",
            },
            "latest_redis_consumer": {
                "cmd": "/Users/jean.machado/projects/PythonSearch/python_search/events/latest_used_entries.py consume",
                "log": "/tmp/log_consumer_redis",
            },
            "spark_consumer": {
                "cmd": "sleep 12;  log_command.sh spark-submit \
        --packages org.apache.spark:spark-sql-kafka-0-10_2.12:3.0.1 \
        $HOME/projects/PythonSearch/python_search/events/consumer.py consume_search_run_performed",
                "log": "/tmp/log_spark",
            },
        }

    def all(self):
        self.run_service("web_api", follow_logs=False)
        self.run_service("zookeeper", follow_logs=False)
        self.run_service("kafka", follow_logs=False)
        self.run_service("redis", follow_logs=False)
        self.run_service("spark_consumer", follow_logs=False)
        self.run_service("latest_redis_consumer", follow_logs=False)

        send_notification("Infra setup finished")

    def run_service(self, service_name, follow_logs=True):
        if not service_name in self._RUNNING_INFO:
            raise Exception("Could not find service: " + service_name)

        print("Starting service: " + service_name)
        cmd = f"LOG_FILE={self._RUNNING_INFO[service_name]['log']} log_command.sh {self._RUNNING_INFO[service_name]['cmd']}"

        self._RUNNING_INFO[service_name]["pid"] = self._run_pid(cmd)

        if follow_logs:
            self.follow_logs(service_name)
        return self.status()

    def _run_pid(self, cmd):
        p = Popen(cmd, shell=True)
        return p.pid

    def status(self, json=False):
        for name, data in self._RUNNING_INFO.items():
            self._RUNNING_INFO[name]["already_run"] = os.path.exists(data["log"])

        if json:
            import json

            return json.dumps(self._RUNNING_INFO)

        return self._RUNNING_INFO

    def status_window(self):
        from python_search.interpreter.cmd import CmdInterpreter

        os.system(f" echo '{self.status(json=True)}' > /tmp/infra_status")
        CmdInterpreter(
            {"cli_cmd": f"  cat /tmp/infra_status | jq . "}
        ).interpret_default()

    def follow_logs(self, service):
        from python_search.interpreter.cmd import CmdInterpreter

        CmdInterpreter(
            {"cli_cmd": f"tail +1f {self._RUNNING_INFO[service]['log']}"}
        ).interpret_default()


def main():
    import fire

    fire.Fire(Infrastructure)


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Optional, Literal


class Browser:
    """
    Abstracts the browser features cross-platform
    """

    BROSERS = Literal["firefox", "chrome"]

    def open(self, url: Optional[str] = None, app_mode=False, incognito=False) -> None:
        """
        performs the open
        """
        params = locals()
        del params["self"]
        cmd_to_run = self.open_shell_cmd(**params)
        print("Comand to run:", cmd_to_run)

        import os

        os.system(cmd_to_run)

    def open_shell_cmd(
        self,
        url: Optional[str] = None,
        app_mode=False,
        incognito=False,
        browser: Optional[BROSERS] = None,
    ) -> str:
        """
        Returns the shell command to open the browser
        """
        url_expr = ""
        if url is not None:
            url_expr = f"'{url}'"

        if browser == "chrome":
            cmd = f" open -a 'Google Chrome' {url_expr}"
        else:
            cmd = f" open -a Firefox {url_expr}"

        return cmd


def main():
    import fire

    fire.Fire(Browser)


if __name__ == "__main__":
    main()



from __future__ import annotations

import os
from typing import Any, List, Optional

from python_search.config import ConfigurationLoader, PythonSearchConfiguration
from python_search.infrastructure.performance import timeit
from python_search.logger import setup_inference_logger
from python_search.search.models import PythonSearchMLFlow
from python_search.search.next_item_predictor.inference.input import ModelInput
from python_search.search.next_item_predictor.transform import ModelTransform

logger = setup_inference_logger()


class Inference:
    """
    Performs the search inference on all existing keys in the moment
    """

    PRODUCTION_RUN_ID = "239c1788635344c0acf98b8c8f26fff2"

    def __init__(
        self,
        configuration: Optional[PythonSearchConfiguration] = None,
        run_id: Optional[str] = None,
        model: Optional[Any] = None,
    ):

        self.debug = os.getenv("DEBUG", False)
        self.run_id = run_id if run_id else self.PRODUCTION_RUN_ID
        if "FORCE_RUN_ID" in os.environ:
            self.run_id = os.environ["FORCE_RUN_ID"]

        if model:
            logger.info("Using custom passed _model")
        else:
            logger.info("Using run id: " + self.run_id)
        configuration = (
            configuration if configuration else ConfigurationLoader().load_config()
        )
        # previous key should be setted in runtime
        self.previous_key = None
        self.all_keys = configuration.commands.keys()
        self._transform = ModelTransform()

        try:
            self.model = model if model else self._load_mlflow_model(run_id=self.run_id)
        except Exception as e:
            print("Failed to load mlflow model")
            self.model = None
            raise e

    @timeit
    def get_ranking(self, predefined_input: Optional[ModelInput] = None) -> List[str]:
        """
        Gets the search from the next item _model
        """
        logger.info("Number of existing keys for inference: " + str(len(self.all_keys)))
        inference_input = (
            predefined_input
            if predefined_input
            else ModelInput.with_given_keys(
                self._transform.inference_embeddings.get_recent_key_with_embedding(),
                self._transform.inference_embeddings.get_recent_key_with_embedding(
                    second_recent=True
                ),
            )
        )
        logger.info("Inference input: " + str(inference_input.__dict__))

        X = self._transform.transform_single(inference_input, self.all_keys)
        Y = self._predict(X)
        result = list(zip(self.all_keys, Y))
        result.sort(key=lambda x: x[1], reverse=True)

        only_keys = [entry[0] for entry in result]
        logger.info("Ranking inference succeeded")

        # logger.debug("Only keys: ", only_keys)
        return only_keys

    @timeit
    def _predict(self, X):
        return self.model.predict(X)

    @timeit
    def _load_mlflow_model(self, run_id=None):
        model = PythonSearchMLFlow().get_next_predictor_model(run_id=run_id)
        return model



from __future__ import annotations

from enum import Enum

from python_search.interpreter.file import FileInterpreter
from python_search.interpreter.url import UrlInterpreter


class EntryType(str, Enum):
    URL = "Url"
    FILE = "File"
    CMD = "Cmd"
    SNIPPET = "Snippet"
    CALLABLE = "Callable"

    @staticmethod
    def all():
        return {
            0: EntryType.URL,
            1: EntryType.SNIPPET,
            2: EntryType.CMD,
            3: EntryType.FILE,
            4: EntryType.CALLABLE,
        }

    @staticmethod
    def from_categorical(categorical: int) -> "EntryType":
        return EntryType.all()[categorical]

    @staticmethod
    def to_categorical(type: "EntryType") -> int:
        """
        @ todo see if we can use self for this function instead, how does it work with enums?

        :param type:
        :return:
        """
        all = list(EntryType.all().values())

        return all.index(type)


def infer_default_type(content: str) -> EntryType:

    if UrlInterpreter.is_url(content):
        return EntryType.URL

    if FileInterpreter.file_exists(content):
        return EntryType.FILE

    return EntryType.SNIPPET



"""

Core entities reused in the hole project.
They should be side-effect free

"""
from __future__ import annotations

from typing import List, Optional

from numpy import ndarray
from typed_pyspark import Dataframe


class Entry:
    """
    An python dictionary we write in PythonSearch
    """

    name: str
    value: Optional[dict]
    embedding: Optional[ndarray]
    similarity_score: Optional[float]

    def __init__(self, *, name: str, value: dict = None, embedding=None):
        self.name = name
        self.value = value
        self.embedding = embedding

    def has_embedding(self) -> bool:
        return self.embedding is not None

    def get_similarity_score(self) -> float:
        """returns a score, if none will then return 0"""
        return self.similarity_score if self.similarity_score else 0.0

    def serialize(self) -> str:
        return f"{self.name} {self.value}"


class Ranking:
    entries: List[Entry] = []

    def __init__(self, *, ranked_entries: List[Entry]):
        self.entries = ranked_entries

    def get_only_names(self) -> List[str]:
        return [entry.name for entry in self.entries]


class InvertedIndex:
    """
    The entity that gets persisted to disk
    """

    entries: List[Entry]

    @staticmethod
    def from_entries_dict(dict: dict) -> InvertedIndex:
        instance = InvertedIndex()
        instance.entries = []
        for key, value in dict.items():
            instance.entries.append(Entry(name=key, value=value))

        return instance

    def serialize(self) -> str:
        pass


Entries = Dataframe(schema={"key": "String", "position": "Integer"})
EntriesType = Entries.type_annotation()



import os


def send_notification(message: str):
    from python_search.environment import is_mac

    """Sends a system notification and sanitizes in case of special chars"""
    clean = message.replace("'", "")

    cmd = f"notify-send '{clean}'"
    if is_mac():
        cmd = f"""osascript -e 'display notification "{clean}"'"""

    os.system(cmd)


def main():
    import fire

    fire.Fire(send_notification)


if __name__ == "__main__":
    main()



from python_search.apps.window_manager import WindowManager
from python_search.entries_group import EntriesGroup
from python_search.environment import is_mac
from python_search.shortcut.gnome import Gnome
from python_search.shortcut.i3 import I3
from python_search.shortcut.mac import Mac


class ShortcutGenerator:
    """
    Generate shortcuts for python search
    """

    def __init__(self, configuration: EntriesGroup):
        self.configuration = configuration
        self.mac = Mac(self.configuration)
        self.gnome = Gnome(self.configuration)
        self.i3 = I3(self.configuration)

    def generate(self):
        if is_mac():
            self.mac.generate()
            return

        if WindowManager.is_gnome():
            self.gnome.generate()
            return

        if WindowManager.is_i3():
            self.i3.generate()
            return



import json
import os

from python_search.logger import setup_data_writter_logger


class GenericDataCollector:
    """
    A generic data writer component that works tightly integrated with spark
    """

    BASE_DATA_DESTINATION_DIR = os.environ["HOME"] + "/.data/data_collection"

    @staticmethod
    def initialize():
        import fire

        return fire.Fire(GenericDataCollector())

    def write(self, *, data: dict, table_name: str, date=None):
        self.logger = setup_data_writter_logger(table_name)
        from datetime import datetime

        datetime.now().timestamp()

        import os

        os.system(
            f"mkdir -p {GenericDataCollector.BASE_DATA_DESTINATION_DIR}/{table_name}"
        )
        import datetime

        file_name = f"{self.data_location(table_name)}/{datetime.datetime.now(datetime.timezone.utc).timestamp()}.json"

        with open(file_name, "w") as f:
            f.write(json.dumps(data))

        self.logger.info(f"File {file_name} written successfully with data {data}")

    def data_location(self, table_name) -> str:
        return f"{GenericDataCollector.BASE_DATA_DESTINATION_DIR}/{table_name}"

    def dataframe(self, table_name):
        from pyspark.sql import DataFrame
        from pyspark.sql.session import SparkSession

        spark = SparkSession.builder.getOrCreate()
        result: DataFrame = spark.read.json(GenericDataCollector().data_location(table_name))
        return result

    def show_data(self, table_name):
        return self.dataframe(table_name).show()


if __name__ == "__main__":
    GenericDataCollector.initialize()



import numpy as np
from pydantic import BaseModel


max_chars_from_key = 30
max_chars_of_body = 100
maxlen = max_chars_of_body + max_chars_from_key


class EntryKeyGeneratorCmd(BaseModel):
    content: str
    temperature: float = 0.2


class DescriptionGenerator:

    RUN_ID = "054613ec285f4b1c86ee81de98b08d06"

    def __init__(self):
        try:
            from python_search.search.models import PythonSearchMLFlow

            self._model = PythonSearchMLFlow().get_entry_description_geneartor(
                run_id=DescriptionGenerator.RUN_ID
            )

            self._chars = PythonSearchMLFlow().get_entry_description_geneartor_dict(
                run_id=DescriptionGenerator.RUN_ID
            )
            self._char_indices = dict(
                (char, self._chars.index(char)) for char in self._chars
            )
        except Exception as e:
            print("Could not load descritpion generator {}".format(e))
            return

    def generate(self, cmd: EntryKeyGeneratorCmd):

        result = ""
        for i in range(0, max_chars_from_key):
            text = cmd.content
            encoded = self._transform(text)
            preds = self._model.predict(encoded)

            next_index = self.sample(preds[0], cmd.temperature)
            next_char = self._chars[next_index]
            text = text + next_char
            result += next_char
        return result

    def sample(self, preds, temperature=1.0):
        preds = np.asanyarray(preds).astype("float64")
        preds = np.log(preds) / temperature

        exp_preds = np.exp(preds)
        preds = exp_preds / np.sum(exp_preds)
        probas = np.random.multinomial(1, preds, 1)
        return np.argmax(probas)

    def _transform(self, text):

        text_len = len(text)
        missing_spaces = text_len < (maxlen - 3)

        if missing_spaces:
            text = " " * missing_spaces

        text = text + " = "

        encoded = np.zeros((1, maxlen, len(self._chars)))

        for i, char in enumerate(text):
            encoded[0, i, self._char_indices[char]] = 1

        return encoded


if __name__ == "__main__":
    import fire

    fire.Fire(DescriptionGenerator)



import os


class Terminal:
    """
    Terminal abstraction for Python Search
    The underlying terminal is kitty but we could support more in the future if needed
    """

    # these parameters are applied both to all kitty windows of pythons search
    # including the generic params and the python search main window
    GLOBAL_TERMINAL_PARAMS = " -o remember_window_size=yes -o placement_strategy=center -o confirm_os_window_close=0  -o remember_window_size=n -o macos_quit_when_last_window_closed=yes "

    # this parameters are applied to cli_cmd terminals

    DEFAULT_HEIGHT = 500
    DEFAULT_WIDTH = 900
    GENERIC_TERMINAL_PARAMS = f" {GLOBAL_TERMINAL_PARAMS} -o initial_window_width={DEFAULT_WIDTH} -o initial_window_height={DEFAULT_HEIGHT} -o font_size=13 "

    def wrap_cmd_into_terminal(
        self, cmd, title=None, hold_terminal_open_on_end=True
    ) -> str:
        """
        wraps the command in a terminal but does not execute it
        """
        shell = (
            os.environ["PYTHON_SEARCH_CUSTOM_SHELL"]
            if "PYTHON_SEARCH_CUSTOM_SHELL" in os.environ
            else "bash"
        )
        shell = "/bin/bash"
        cmd = f'{shell} -c "{cmd}"'

        hold = ""
        if hold_terminal_open_on_end:
            hold = " --hold "

        final_cmd = (
            f'kitty {hold} {Terminal.GENERIC_TERMINAL_PARAMS} -T "{title}" {cmd} '
        )

        return final_cmd



from typing import List, Tuple
import random
import sys
import numpy as np
from python_search.config import ConfigurationLoader
from keras import layers
import keras
from python_search.search.next_item_predictor.mlflow_logger import configure_mlflow


def extract_string_from_entries(entries) -> List[str]:
    X = []
    for key, entry in entries.items():
        if "url" in entry:
            item = str(entry["url"])
        if "snippet" in entry:
            item = str(entry["snippet"])
        if "file" in entry:
            item = str(entry["file"])
        if "callable" in entry:
            item = str(entry["callable"])
        if "cmd" in entry:
            item = str(entry["cmd"])
        if "cli_cmd" in entry:
            item = str(entry["cli_cmd"])

        X.append(item)

    return X


def sample(preds, temperature=1.0):
    preds = np.asanyarray(preds).astype("float64")
    preds = np.log(preds) / temperature

    exp_preds = np.exp(preds)
    preds = exp_preds / np.sum(exp_preds)
    probas = np.random.multinomial(1, preds, 1)
    return np.argmax(probas)


def run(epochs=50):
    # create dataset
    entries = ConfigurationLoader().load_entries()
    result = extract_string_from_entries(entries)
    dataset = list(zip(result, entries.keys()))

    filtered_dataset: List[Tuple[str, str]] = []
    for value, key in dataset:
        if key.startswith("no key"):
            continue
        filtered_dataset.append([value, key])

    sentences = []
    next_char = []
    max_chars_from_key = 30
    max_chars_of_body = 100
    maxlen = max_chars_of_body + max_chars_from_key

    for value, key in filtered_dataset:
        for i in range(0, len(key)):
            if i > (max_chars_from_key - 3):
                break

            sentence = value[0:100] + " = " + key[0:(i)]

            current_size = len(sentence)
            if current_size < maxlen:
                sentence = (" " * (maxlen - current_size)) + sentence

            sentences.append(sentence)
            next_char.append(key[i])

    for i, sentence in enumerate(sentences[0:100]):
        print(f"sentence: {sentence}, next_char = {next_char[i]} ")

    text = " ".join(next_char) + " ".join(sentences)

    chars = sorted(list(set(text)))
    char_indices = dict((char, chars.index(char)) for char in chars)

    chars[0:5]

    x = np.zeros((len(sentences), maxlen, len(chars)), dtype=np.bool)
    y = np.zeros((len(sentences), len(chars)), dtype=np.bool)

    for i, sentence in enumerate(sentences):
        # print(f"Sentence: {sentence}")
        for t, char in enumerate(sentence):
            # print(t)

            x[i, t, char_indices[char]] = 1

        y[i, char_indices[next_char[i]]] = 1

    model = keras.models.Sequential()
    model.add(layers.LSTM(128, input_shape=(maxlen, len(chars))))
    model.add(layers.Dense(len(chars), activation="softmax"))
    optimizer = keras.optimizers.RMSprop(lr=0.011)
    model.compile(loss="categorical_crossentropy", optimizer=optimizer)

    # train model
    epoch = 0
    for epoch in range(1, epochs):
        print("\n => Epoch: ", epoch)
        model.fit(x, y, batch_size=64, epochs=1)
        start_index = random.randint(0, len(text) - maxlen - 1)
        generated_text = text[start_index : start_index + maxlen]
        print(' -- generating with seed: "' + generated_text + '"')

        for temperature in [0.2, 1.2]:
            print(f"\n ----  Temperature: {temperature}, Input text follows: ")
            sys.stdout.write(generated_text)
            print("\n Generated text follows: ")
            for i in range(30):
                sampled = np.zeros((1, maxlen, len(chars)))
                for t, char in enumerate(generated_text):
                    sampled[0, t, char_indices[char]] = 1

                preds = model.predict(sampled, verbose=0)[0]
                next_index = sample(preds, temperature)
                next_char = chars[next_index]

                generated_text += next_char
                generated_text = generated_text[1:]

                sys.stdout.write(next_char)

    mlflow = configure_mlflow(experiment_name="entry_description_generator")
    mlflow.start_run()
    mlflow.log_dict(chars, "chars")
    mlflow.log_param("epochs", epoch)
    mlflow.keras.log_model(model, "model")
    mlflow.end_run()


if __name__ == "__main__":
    import fire

    fire.Fire()



from __future__ import annotations

import logging
from collections import namedtuple
from typing import List, Literal, Optional

from python_search.config import PythonSearchConfiguration
from python_search.events.latest_used_entries import RecentKeys
from python_search.events.ranking_generated import (
    RankingGenerated,
    RankingGeneratedWriter,
)
from python_search.feature_toggle import FeatureToggle
from python_search.infrastructure.performance import timeit
from python_search.search.ranked_entries import RankedEntries
from python_search.search.results import FzfOptimizedSearchResults

ModelInfo = namedtuple("ModelInfo", "features label")


class Search:
    """
    Generates the search for python search
    """

    NUMBER_OF_LATEST_ENTRIES = 7

    _model_info = ModelInfo(["position", "key_lenght"], "input_lenght")
    _inference = None

    def __init__(self, configuration: Optional[PythonSearchConfiguration] = None):
        self._configuration = configuration
        self._feature_toggle = FeatureToggle()
        self._model = None
        self._entries_result = FzfOptimizedSearchResults()
        self._entries = None
        self._ranking_generator_writer = RankingGeneratedWriter()
        self._ranking_method_used: Literal[
            "RankingNextModel", "BaselineRank"
        ] = "BaselineRank"

        if self._feature_toggle.is_enabled("ranking_next"):
            from python_search.search.next_item_predictor.inference.inference import (
                Inference,
            )

            try:
                self._inference = Inference(self._configuration)
            except Exception as e:
                print(
                    f"Could not initialize the inference component. Proceeding without inference, details: {e}"
                )

    @timeit
    def search(self) -> str:
        """
        Recomputes the rank and saves the results on the file to be read
        """

        self._entries: dict = self._configuration.commands
        # by default the rank is just in the order they are persisted in the file
        self._ranked_keys: List[str] = list(self._entries.keys())

        if self._feature_toggle.is_enabled("ranking_next") and self._inference:
            self._rerank_via_model()

        """Populate the variable used_entries  with the results from redis"""
        result = self._merge_with_latest_used()

        ranknig_generated_event = RankingGenerated(
            ranking=[i[0] for i in result[0:100]]
        )
        self._ranking_generator_writer.write(ranknig_generated_event)
        print(f"Ranking generated UUID {ranknig_generated_event.uuid}")
        result_str = self._entries_result.build_entries_result(
            result, ranknig_generated_event.uuid
        )

        return result_str

    def _rerank_via_model(self):
        try:
            self._ranked_keys = self._inference.get_ranking()
            self._ranking_method_used = "RankingNextModel"
        except Exception as e:

            print(f"Failed to perform inference, reason {e}")

            # raise e

    def _merge_with_latest_used(self) -> RankedEntries.type:
        """
        Merge the search with the latest entries
        """

        result = []

        latest_entries = self._fetch_latest_entries()

        if latest_entries:
            for key in latest_entries:
                if key not in self._entries:
                    # key not found in _entries
                    continue

                content = self._entries[key]

                # sometimes there can be a bug of saving something other than dicts as _entries
                if type(content) != dict:
                    logging.warning(f"Entry is not a dict {content}")
                    continue

                content["tags"] = content.get("tags", []) + ["RecentlyUsed"]
                result.append((key, content))
                # delete key
                self._ranked_keys.remove(key)

        for key in self._ranked_keys:
            if key not in self._entries:
                # key not found in _entries
                continue

            entry = self._entries[key]
            if type(entry) == dict:
                existing_tags = entry.get("tags", [])
                if type(existing_tags) == str:
                    existing_tags = [existing_tags]

                entry["tags"] = existing_tags + [self._ranking_method_used]
            result.append((key, entry))

        # the result is the one to be returned, final_key_list is to be used in the cache
        return result

    def _fetch_latest_entries(self):
        """Populate the variable used_entries  with the results from redis"""
        if not self._feature_toggle.is_enabled("ranking_latest_used"):
            return

        entries = RecentKeys().get_latest_used_keys()

        # only use the latest 7 entries for the top of the search
        return entries[: self.NUMBER_OF_LATEST_ENTRIES]


if __name__ == "__main__":
    import fire

    fire.Fire()



from typing import List

from python_search.config import ConfigurationLoader


class EntriesLoader:
    """Class to access the current existing key"""

    @staticmethod
    def load_all_keys() -> List[str]:
        keys = list(ConfigurationLoader().load_entries().keys())

        print("Loaded in total " + str(len(keys)) + " keys")

        return keys



from __future__ import annotations

import shutil
from datetime import datetime

from python_search.config import PythonSearchConfiguration
from python_search.events.run_performed import RunPerformed


class FilesystemEntryInserter:
    """
    Class responsible to the actual writing of the new entry in the filesystem
    """

    NEW_ENTRIES_STRING = "# NEW_ENTRIES_HERE"

    def __init__(self, configuration: PythonSearchConfiguration):
        self._configuration = configuration
        self._file_to_append = (
            self._configuration.get_project_root() + "/entries_main.py"
        )
        self._new_entries_string = FilesystemEntryInserter.NEW_ENTRIES_STRING

    def insert(self, key: str, entry: dict):

        entry["created_at"] = datetime.now().isoformat()

        if "tags" in entry and self._configuration.tags_dependent_inserter_marks:
            for tag in entry["tags"]:
                if tag in self._configuration.tags_dependent_inserter_marks.keys():
                    self._new_entries_string = (
                        self._configuration.tags_dependent_inserter_marks[tag][0]
                    )
                    self._file_to_append = (
                        self._configuration.get_project_root()
                        + "/"
                        + self._configuration.tags_dependent_inserter_marks[tag][1]
                    )

        row_entry = str(entry)
        line_to_add = f"    '{key}': {row_entry},"
        self._append_entry(line_to_add)

        from python_search.events.run_performed.writer import LogRunPerformedClient
        LogRunPerformedClient().send(
            RunPerformed(key=key, query_input="", shortcut=False)
        )

        from python_search.apps.notification_ui import send_notification

        send_notification(f"Entry {row_entry} inserted successfully")

    def _append_entry(self, line_to_add: str):
        """
        This script does the following:
            Copies the main file,
            add the new entry to it
            Compile to see if it is still valid python
            If so, then replaces it
        """
        copy_file = self._file_to_append + "cpy"

        print(f"Copying file as backup {self._file_to_append} => {copy_file}")
        shutil.copyfile(self._file_to_append, copy_file)

        with open(copy_file, "w") as out, open(self._file_to_append) as source_file:
            for line in source_file:
                out.write(line)
                if self._new_entries_string in line:
                    # insert text.
                    print(f"Writing line: {line_to_add}")
                    out.write(line_to_add + "\n")

        # compile and make sure the file is a valid python
        import os

        if os.system(f"python3 -m compileall -q {copy_file}") != 0:
            message = "Copy of file does not compile so wont proceed replacing!"
            from python_search.apps.notification_ui import send_notification

            send_notification(message)
            raise Exception(message)

        shutil.move(copy_file, self._file_to_append)



from typing import Any

import matplotlib.pyplot as plt
import xgboost as xgb
from sklearn.metrics import mean_absolute_error

Model = Any


class TrainXGBoost:
    def train(self, X_train, X_test, Y_train, Y_test) -> Model:

        model = xgb.XGBRegressor(
            tree_method="hist",
            eval_metric=mean_absolute_error,
            objective="reg:squarederror",
            eval_set=[(X_train, Y_train), (X_test, Y_test)],
            booster="gbtree",
            verbose=True,
        )

        eval_set = [(X_train, Y_train), (X_test, Y_test)]
        model.fit(X_train, Y_train, eval_set=eval_set)

        print("Xgboost was trained successfully!")
        return model

    def _evaluate_metrics(self, model, display_metrics=False):

        results = model.evals_result()
        epochs = len(results["validation_0"]["rmse"])
        x_axis = range(0, epochs)
        fig, ax = plt.subplots()
        ax.plot(x_axis, results["validation_0"]["rmse"], label="Train")
        ax.plot(x_axis, results["validation_1"]["rmse"], label="Test")
        ax.legend()
        plt.ylabel("rmse")
        plt.title("XGBoost_training rmse")
        if display_metrics:
            plt.show()

        results = model.evals_result()
        epochs = len(results["validation_0"]["mean_absolute_error"])
        x_axis = range(0, epochs)
        fig, ax = plt.subplots()
        ax.plot(x_axis, results["validation_0"]["mean_absolute_error"], label="Train")
        ax.plot(x_axis, results["validation_1"]["mean_absolute_error"], label="Test")
        ax.legend()
        plt.ylabel("mae")
        plt.title("XGBoost_training mae")
        if display_metrics:
            plt.show()



from fastapi import FastAPI
from fastapi.responses import PlainTextResponse

from python_search.entry_description_generator.description_geneartor import (
    DescriptionGenerator,
    EntryKeyGeneratorCmd,
)
from python_search.entry_type.classifier_inference import (
    EntryData,
    PredictEntryTypeInference,
)
from python_search.events.latest_used_entries import RecentKeys
from python_search.events.run_performed import RunPerformed
from python_search.events.run_performed.writer import RunPerformedWriter
from python_search.config import ConfigurationLoader
from python_search.search.search import Search

import pyroscope

pyroscope.configure(
    application_name="my.python.app",  # replace this with some name for your application
    server_address="http://host.docker.internal:4040",  # replace this with the address of your pyroscope server
    sample_rate=1,  # default is 100
)
PORT = 8000

app = FastAPI()

generator = Search(ConfigurationLoader().load_config())
ranking_result = generator.search()
entry_type_inference = PredictEntryTypeInference()
description_generator = DescriptionGenerator()


def reload_ranking():
    with pyroscope.tag_wrapper({"endpoint": "reload_ranking"}):
        global generator
        global ranking_result
        ranking_result = Search(ConfigurationLoader().reload()).search()
        return ranking_result


@app.get("/ranking/generate", response_class=PlainTextResponse)
def generate_ranking():
    global ranking_result
    return ranking_result


@app.get("/ranking/reload", response_class=PlainTextResponse)
def reload():
    reload_ranking()


@app.get("/ranking/reload_and_generate", response_class=PlainTextResponse)
def reload():
    return reload_ranking()


@app.get("/_health")
def health():
    global generator

    from python_search.events.latest_used_entries import RecentKeys

    entries = RecentKeys().get_latest_used_keys()
    run_id = None
    if generator._inference is not None:
        run_id = generator._inference.PRODUCTION_RUN_ID

    return {
        "keys_count": len(ConfigurationLoader().load_config().commands.keys()),
        "run_id": run_id,
        "latest_used_entries": entries,
        "initialization_time": ConfigurationLoader().load_config()._initialization_time,
    }


@app.post("/log_run")
def log_run(event: RunPerformed):

    RecentKeys.add_latest_used(event.key)
    RunPerformedWriter().write(event)
    # regenerate the search after running a key
    reload()

    return event


@app.post("/entry_type/classify")
def predict_entry_type_endpoint(entry: EntryData):
    type, uuid = entry_type_inference.predict_entry_type(entry)
    return {"predicted_type": type, "prediction_uuid": uuid}


@app.post("/entry/generate_description")
def generate_description(entry: EntryKeyGeneratorCmd):
    with pyroscope.tag_wrapper({"endpoint": "generate_description"}):
        result = description_generator.generate(entry)
        return {"generated_description": result}


@app.get("/recent_history")
def recent_history_endpoint():
    return {"history": RecentKeys().get_latest_used_keys()}


def main():

    import os

    import uvicorn

    os.putenv("WEB_CONCURRENCY", "0")
    reload = False
    if "PS_DEBUG" in os.environ:
        print("Debug mode is ON, enabling reload")
        reload = True
    else:
        print("Debug mode is OFF")

    uvicorn.run("python_search.web_api:app", host="0.0.0.0", port=PORT, reload=reload)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['python_search',
 'python_search.apps',
 'python_search.data_ui',
 'python_search.entry_capture',
 'python_search.entry_description_generator',
 'python_search.entry_type',
 'python_search.events',
 'python_search.events.ranking_generated',
 'python_search.events.run_performed',
 'python_search.infrastructure',
 'python_search.init',
 'python_search.interpreter',
 'python_search.plugins',
 'python_search.sdk',
 'python_search.search',
 'python_search.search.next_item_predictor',
 'python_search.search.next_item_predictor.features',
 'python_search.search.next_item_predictor.features.entry_embeddings',
 'python_search.search.next_item_predictor.inference',
 'python_search.search_ui',
 'python_search.shortcut']

package_data = \
{'': ['*']}

install_requires = \
['PySimpleGUI>=4.60.1,<5.0.0',
 'PyYAML>=6.0,<7.0',
 'certifi>=2022.6.15,<2023.0.0',
 'colorama>=0.4.5,<0.5.0',
 'dill>=0.3.5.1,<0.4.0.0',
 'fire>=0.4.0,<0.5.0',
 'pydantic>=1.9.1,<2.0.0',
 'pyroscope-io>=0.8.0,<0.9.0']

extras_require = \
{':extra == "server"': ['mlflow>=1.29.0,<1.30.0'],
 'server': ['fastapi>=0.79.0,<0.80.0',
            'kafka-python>=2.0.2,<3.0.0',
            'scipy>=1.8.1',
            'xgboost>=1.6.1,<2.0.0',
            'matplotlib>=3.5.2,<4.0.0',
            'pandas>=1.4.3,<2.0.0',
            'redis>=4.3.4,<5.0.0',
            'uvicorn>=0.18.2,<0.19.0',
            'numpy>=1.23.4',
            'msgpack-numpy>=0.4.8',
            'pyspark>=3.0',
            'keras>=2.10',
            'tensorflow>=2.10.0',
            'arize>=5.2.0,<6.0.0',
            'jupyterlab>=3.5.0,<4.0.0']}

entry_points = \
{'console_scripts': ['browser = python_search.apps.browser:main',
                     'clipboard = python_search.apps.clipboard:main',
                     'collect_input = python_search.apps.collect_input:main',
                     'feature_toggle = '
                     'python_search.feature_toggle:FeatureToggle.main',
                     'generic_data_collector = '
                     'python_search.data_collector:GenericDataCollector.initialize',
                     'next_item_pipeline = '
                     'python_search.search.next_item_predictor.pipeline:main',
                     'notify_send = python_search.apps.notification_ui:main',
                     'ps_container = python_search.container:start',
                     'ps_webapi = python_search.sdk.web_api_sdk:main',
                     'python_search = python_search.cli:main',
                     'python_search_infra = '
                     'python_search.infrastructure.infrastructure:main',
                     'python_search_webapi = python_search.web_api:main',
                     'reminders = python_search.plugins.reminders:main',
                     'run_entry = python_search.cli:_run_key_bin']}

setup_kwargs = {
    'name': 'python-search',
    'version': '0.9.6',
    'description': 'Build your knowledge database in python and retrieve it efficiently',
    'long_description': 'None',
    'author': 'Jean Carlo Machado',
    'author_email': 'machado.c.jean@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from __future__ import annotations

import os
import subprocess

import streamlit as st


def restart_app():
    result = subprocess.check_output("pkill streamlit", shell=True, text=True)
    st.write(f"Result: {result}")


import streamlit as st


def check_password():
    """Returns `True` if the user had the correct password."""

    if "PS_DISABLE_PASSWORD" in os.environ:
        print("Password disabled")
        return True

    def password_entered():
        import os

        """Checks whether a password entered by the user is correct."""
        if st.session_state["password"] == os.environ["PS_WEB_PASSWORD"]:
            st.session_state["password_correct"] = True
            del st.session_state["password"]  # don't store password
        else:
            st.session_state["password_correct"] = False

    if "password_correct" not in st.session_state:
        # First run, show input for password.
        st.text_input(
            "Password", type="password", on_change=password_entered, key="password"
        )
        return False
    elif not st.session_state["password_correct"]:
        # Password not correct, show input + error.
        st.text_input(
            "Password", type="password", on_change=password_entered, key="password"
        )
        st.error("😕 Password incorrect")
        return False
    else:
        # Password correct.
        return True
